#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

"""
Block Properties:

1. Blocks are indexed components that contain other components
   (including blocks)

2. Blocks have a global attribute that defines whether construction is
   deferred.  This applies to all components that they contain except
   blocks.  Blocks contained by other blocks use their local attribute
   to determine whether construction is deferred.

NOTE: Blocks do not currently maintain statistics about the sets,
parameters, constraints, etc that they contain, including these
components from subblocks.
"""

__all__ = ['Block', 'active_components_generator', 'active_subcomponents_generator']
import sys
import weakref

from plugin import *

from pyutilib.component.core import alias, ExtensionPoint
from pyutilib.misc import Container

from component import Component
from sets import Set, _SetContainer
from rangeset import RangeSet
from var import Var
from constraint import Objective, Constraint
from param import Param
from misc import apply_indexed_rule
from set_types import IntegerSet, BooleanSet

from indexed_component import IndexedComponent
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict

import logging
logger = logging.getLogger('coopr.pyomo')

# 
# we now include a number of generators for iterating over various components within a block. 
# the primary purpose is to streamline code, as this kind of iteration is done *everywhere*
# in Coopr. any generally useful iteration schemes should be moved into the following segment.
#

# iterates over all active components (e.g., Var) of the specified type in the input block.
# sorting, if specified, is by component name.
def active_components_generator(a_block, component_type, sort_by_keys=False):
    active_components = a_block.active_components(component_type)
    if sort_by_keys is True:
        for component_name in sorted(active_components): 
            component = active_components[component_name]
            if (not component.active):
                continue
            yield component
    else:
        for component in active_components.itervalues():
            if (not component.active):
                continue
            yield component

# iterates over all active sub_components (e.g., _VarData) of all active components of the 
# specified type in the input block. sorting, if specified, is by component name, and within 
# a component, by index.
def active_subcomponents_generator(a_block, component_type, sort_by_keys=False):
    active_components = a_block.active_components(component_type)
    if sort_by_keys is True:
        for component_name in sorted(active_components):
            component = active_components[component_name]
            if (not component.active):
                continue
            for index in sorted(component._data):
                sub_component = component[index]
                if (not sub_component.active):
                    continue
                yield sub_component
    else:
        for component in active_components.itervalues():
            if (not component.active):
                continue
            for sub_component in component.itervalues():
                if (not sub_component.active):
                    continue
                yield sub_component

# 
#
#

class BlockComponents(object):

    #
    # IMPT: Default __getstate__ and __setstate__ are fine - this class owns no weakrefs or slots.
    #

    def __init__(self):
        #
        # Define component dictionary: component type -> instance
        #
        self._component={}
        for item in ModelComponentFactory.services():
            self._component[  ModelComponentFactory.get_class(item) ] = OrderedDict()
        #
        # A list of the declarations, in the order that they are
        # specified.
        #
        self._declarations=OrderedDict()

    def __getitem__(self, name):
        if isinstance(name,basestring):
            return self._declarations.get(name, None)
        return self._component.get(name, None)

    def _clear_attribute(self, name):
        val = self[name]
        if val is None:
            return

        if (type(val) == weakref and isinstance(val(), Block)):
            val().__dict__['_parent'] = None
            val()._model = weakref.ref(val)
        elif isinstance(val, Block):
            val.__dict__['_parent'] = None
            val._model = weakref.ref(val)
        else:
            val._model = None
            val._parent = None

        del self._component[ val.type() ][name]
        del self._declarations[name]
        self.__dict__[name]=None

    def _add_component(self, name, val):
        self._component[val.type()][name]=val
        self._declarations[name] = val
        self.__dict__[name]=val

    def __setattr__(self,name,val):
        #
        # Try to set the value. This may fail if the attribute
        # does not already exist in this model, if the set_value
        # function is not defined, or if a bad value is
        # provided. In the latter case, a ValueError will be
        # thrown, which we raise. Otherwise, this is an object
        # that we need to set directly.
        #
        #try:
        #    self.__dict__[name].set_value(val)
        #except ValueError, e:
        #    raise
        #except AttributeException, e:
        #    self.__dict__[name]=val
        if name in self.__dict__:
            if val is None:
                del self.__dict__[name]
                return
            
            setter = getattr(self.__dict__[name], 'set_value', None)
            if setter is None:
                if isinstance(self.__dict__[name], Component):
                    raise Exception(
                        "Expected component %s to have a 'set_value' "
                        "method, but none was found." % \
                        ( self.__dict__[name],))
                self.__dict__[name]=val
            else:
                setter(val)
        else:
            self.__dict__[name]=val

    def model(self):
        """Return the top-level block that defines this model."""
        return self._model()

    def keys(self):
        return self._declarations.keys()

    def __iter__(self):
        return self._declarations.iterkeys()

    def iterkeys(self):
        return self._declarations.iterkeys()

    def itervalues(self):
        return self._declarations.itervalues()

    def iteritems(self):
        return self._declarations.iteritems()

    def contains_component(self, ctype):
        return ctype in self._component

    def components(self, ctype=None):
        """
        Return information about the block components.  If ctype is None, return the dictionary
        that maps {component type -> {name -> instance}}.  Otherwise, return the dictionary
        that maps {name -> instance} for the specified component type.
        """
        if ctype is None:
            return self._component
        if ctype in self._component:
            return self._component[ctype]
        # FIXME: is this really an exception?  Shouldn't it return {}?
        raise KeyError, "Unknown component type: %s" % str(ctype)

    def active_components(self, _ctype=None):
        """
        Returns the active components in this block.  If _ctype is None, return the
        dictionary that maps {component type -> {name -> instance}}.  Otherwise, return
        the dictionary that maps {name -> instance} for the specified component type.
        """
        tmp = {}
        if _ctype is None:
            for ctype in self._component:
                tmp[ctype]=OrderedDict()
        elif _ctype in self._component:
            tmp[_ctype]=OrderedDict()
        else:
            raise KeyError, "Unknown component type: %s" % str(_ctype)
        for ctype in tmp:
            for name in self._component[ctype]:
                comp = self._component[ctype][name]
                if comp.active:
                    tmp[ctype][name] = comp
        if not _ctype is None:
            return tmp[_ctype]
        return tmp

    def is_constructed(self):
        """
        A boolean indicating whether or not all *active* components of the
        input model have been properly constructed.
        """
        component_map = arg.active_components()
        for type, entries in component_map.iteritems():
            for component_name, component in entries.iteritems():
                if component.is_constructed() is False:
                    return False
        return True

    def pprint(self, filename=None, ostream=None, verbose=False):
        """
        Print a summary of the model info
        """
        if ostream is None:
            ostream = sys.stdout
        if filename is not None:
            OUTPUT=open(filename,"w")
            self.pprint(ostream=OUTPUT, verbose=verbose)
            OUTPUT.close()
            return
        if ostream is None:
            ostream = sys.stdout
        #
        # We hard-code the order of the core Pyomo modeling
        # components, to ensure that the output follows the logical order
        # that expected by a user.
        #
        items = [Set, RangeSet, Param, Var, Objective, Constraint, Block]
        for item in ExtensionPoint(IModelComponent):
            if not item in items:
                items.append(item)

        # Currently, model components are not actually being registered
        # with the IModelComponent extension point (1 Nov 2010), so as a
        # workaround, we will loop through the components and add any
        # new components that actually have defined members.
        extra_items = []
        for item, members in self._component.iteritems():
            if item not in items and len(members):
                extra_items.append(item)
        # extra items get added alphabetically (so output is consistent)
        items.extend(sorted(extra_items))

        for item in items:
            if not item in self._component:
                continue
            keys = self._component[item].keys()
            keys.sort()
            #
            # NOTE: these conditional checks should not be hard-coded.
            #
            print >>ostream, len(keys), item.__name__+" Declarations"
            for key in keys:
                self._component[item][key].pprint(ostream=ostream, verbose=verbose)
            print >>ostream, ""
        #
        # Model Order
        #
        print >>ostream, len(self._declarations),"Declarations:",
        for name in self._declarations:
            print >>ostream, name,
        print >>ostream, ""

# 
#
#

class Block(IndexedComponent):
    """
    A block in an optimization model.  By default, this defers construction of
    components until data is loaded.
    """

    alias("Block", "Blocks are indexed components that contain one or more " + \
          "other model components.")

    #
    # IMPT: Must over-ride __getstate__ and __setstate__ - this class owns 
    #       weakrefs via the '_model' and '_parent' attributes.
    #

    def __init__(self, *args, **kwargs):
        """Constructor"""
        self._rule = kwargs.pop('rule', None )
        kwargs.setdefault('ctype', Block)
        IndexedComponent.__init__(self, *args, **kwargs)
        #
        self._constructed = False
        self._defer_construction=True
        # By default, a Block is a top-level (i.e. Model) block
        self._model = weakref.ref(self)
        self._parent = None
        self.components =  BlockComponents()

    def __getstate__(self):
        """ TODO """
        return IndexedComponent.__getstate__(self)

    def __setstate__(self, state):
        """ TODO """
        IndexedComponent.__setstate__(self, state)
        if '_parent' in self.__dict__.keys() and self._parent is not None and type(self._parent) != weakref.ref:
            self._parent = weakref.ref(self._parent)
      
    def all_blocks(self, blockList=None):
        """ TODO """
        ans = blockList or [ self ]
        subBlocks = self.components.components(Block).values()
        ans.extend(subBlocks)
        for block in subBlocks:
            block.all_blocks(ans)
        return ans

    def block_namespace(self):
        """ TODO """
        fullName = ''
        b = self
        while b._parent is not None and b._parent() is not None:
            fullName = b.name + '.' + fullName
            b = b._parent()
        return fullName


    def concrete_mode(self):
        """Configure block to immediately construct components"""
        self._defer_construction=False

    def symbolic_mode(self):
        """Configure block to defer construction of components"""
        self._defer_construction=True

    def components(self, ctype=None):
        """
        Return information about the block components.  If ctype is None, return the dictionary
        that maps {component type -> {name -> instance}}.  Otherwise, return the dictionary
        that maps {name -> instance} for the specified component type.
        """
        return self.components.components(ctype)

    def active_components(self, _ctype=None):
        """
        Returns the active components in this block.  If _ctype is None, return the
        dictionary that maps {component type -> {name -> instance}}.  Otherwise, return
        the dictionary that maps {name -> instance} for the specified component type.
        """
        return self.components.active_components(_ctype)

    def has_discrete_variables(self):
        """ TODO """
        for variable_name, variable in self.active_components(Var).items():
            if isinstance(variable.domain, IntegerSet) or isinstance(variable.domain, BooleanSet):
                return True

        for block in self.components.components(Block).values():
           if block.has_discrete_variables() is True:
              return True
            
        return False

    # flags *all* active variables and their composite _VarData objects as stale.
    # fixed variables are flagged as non-stale. the state of all inactive variables
    # is left unchanged.
    def flag_vars_as_stale(self):

        for block in self.all_blocks():
            active_variables = block.active_components(Var)
            for variable in active_variables.itervalues():
                variable.flag_as_stale()

    def find_component(self, label):
        """ TODO """
        cList = label.split('.')
        obj = self
        while cList:
            c = cList.pop(0).split(':')
            if len(c) > 1:
                idx = c[1].split(',')
                for i, val in enumerate(idx):
                    if val[0]=='#':
                        idx[i] = int(val[1:])
                    elif val[0]=='$':
                        idx[i] = val[1:]
                    elif val[0]=='!':
                        idx[i] = None
                    else:
                        # last-ditch effort to find things
                        try:
                            tmp = int(val)
                            idx[i] = tmp
                        except ValueError:
                            pass
                        #raise ValueError(
                        #    "missing index type specifier for index %s in canonical label %s" % (val,label) )
                if len(idx) > 1:
                    idx = tuple(idx)
                else:
                    idx = idx[0]
            else:
                idx = None
            try:
                obj = getattr(obj, c[0])[idx]
            except AttributeError:
                return None
        return obj

    def construct(self, data=None):
        """ TODO """
        if __debug__ and logger.isEnabledFor(logging.DEBUG):
            logger.debug( "Constructing %s %s",
                          self.__class__.__name__, self.name )
        if self._constructed:
            return
        self._constructed=True
        if self._rule is None:
            # FIXME: The following code needs to be re-established once the "FIXME" 
            #        on line 175 of PyomoModel.py is addressed.
            logger.warn("No construction rule or expression specified for "
                        "%s '%s'", self.__class__.__name__, self.name)
            return

        def _generate_name(idx):
            if type(idx) in (tuple, list):
                return '['+','.join([_generate_name(x) for x in idx])+']'
            else:
                return str(idx)

        for val in self._index:
            name = _generate_name(val)
            if __debug__ and logger.isEnabledFor(logging.DEBUG):
                logger.debug( "   Constructing %s index %s",
                              self.__class__.__name__, str(name) )
            if val is None:
                apply_indexed_rule( self, self._rule, self, () )
                # FIXME: This may be a circular reference, but it's letting us make
                #        non-indexed blocks for now (ZF)
                block = self
            else:
                block = self.__class__()
                self._add_component(name, block)
                apply_indexed_rule( self, self._rule, block, val )
                self._data[val] = block
            for tmp in block.components.itervalues():
                data_to_use = None
                if data is not None:
                    data_to_use = data.get(val,{}).get(tmp.name,None)
                tmp.construct(data_to_use)
        #self._index=constructed_indices

    def _add_temporary_set(self,val):
        """ TODO """
        # NB: only "old" IndexedComponent contain _index_set.
        # SparseIndexedComponent does not have that attribute
        _component_sets = getattr(val, '_index_set', None)
        if _component_sets is None:
            _construct_index_set = False
            _component_sets = getattr(val, '_implicit_subsets', None)
        else:
            _construct_index_set = True
        if _component_sets is not None:
            ctr=0
            for tset in _component_sets:
                if tset.name == "_unknown_":
                    self._construct_temporary_set(
                      tset,
                      val.name+"_index_"+str(ctr)
                    )
                ctr += 1
            if _construct_index_set:
                val._index = self._construct_temporary_set(
                    _component_sets,
                    val.name+"_index" )
        if isinstance(val._index,_SetContainer) and val._index.name == "_unknown_":
            self._construct_temporary_set(val._index,val.name+"_index")
        if getattr(val,'domain',None) is not None and val.domain.name == "_unknown_":
            self._construct_temporary_set(val.domain,val.name+"_domain")

    def _construct_temporary_set(self, obj, name):
        """ TODO """
        if type(obj) is tuple:
            if len(obj) == 1:                #pragma:nocover
                raise Exception, "Unexpected temporary set construction"
            else:
                tobj = obj[0]
                for t in obj[1:]:
                    tobj = tobj*t
                setattr(self,name,tobj)
                tobj.virtual=True
                return tobj
        if isinstance(obj,Set):
            setattr(self,name,obj)
            return obj

    def _clear_attribute(self, name, only_block=False):
        """
        Cleanup the pre-existing model attribute
        """
        #
        # We don't delete here because that would destruct the component, which might
        # be shared with another block.
        #
        self.__dict__[name]=None
        if only_block:
            return

        self.components._clear_attribute(name)

    def _add_component(self, name, val):
        """ TODO """
        if not val.valid_model_component():
            raise RuntimeError, "Cannot add '%s' as a component to a model" % str(type(val))
        self._clear_attribute(name)

        # all Pyomo components have names. 
        val.name=name

        self._add_temporary_set(val)
        self.components._add_component(name,val)
        self.__dict__[name]=val

        # Presumably self._model refers to the 'root' Block that will
        # eventually be solved, whereas '_parent' refers to the
        # immediate parent.

        # Update the (new) child block's parent.
        if isinstance(val, Block):
            if val._parent is not None and val._parent() is not None:
                # Nothing really wrong here, but until we are more
                # careful about checking that this actually works,
                # complain loudly.
                print "WARNING: Reassigning a block attached to a model"
            # NB: use __dict__ directly to prevent infinite recursion
            val.__dict__['_parent'] = weakref.ref(self)
        else:
            val._parent = weakref.ref(self)

        # Update the new component's model pointer
        val._model = self._model

        # Support implicit rule names
        frame = sys._getframe(2)
        locals_ = frame.f_locals
        if getattr(val,'rule',None) is None and val.name+'_rule' in locals_:
            val.rule = locals_[val.name+'_rule']
        if not self._defer_construction:
            val.construct(None)

    def __setattr__(self,name,val):
        """Set attributes"""
        #
        # Set Model Declaration
        #
        if name == '_model':
            if val is not None and not isinstance(val, Block) and \
                    not (type(val) == weakref.ref and isinstance(val(), Block)):
                raise ValueError, "Cannot set Block._model to a non-Block " \
                      "object (%s)" % ( val )
            self.__dict__['_model'] = val
            # Update all child components.  
            #
            # NB: use __dict__.get because Component.__init__() assigns
            # "self._model = None" before Block.__init__ defines the
            # _declarations map.
            #
            # NB: this recursion is critical for pickling to work: when
            # the model is pickled, its __getstate__ reassigns the
            # weakref model pointer to a hard reference, which trickles
            # down to all child components through here.
            components = self.__dict__.get('components', None)
            if not components is None:
                for subcomp in components.itervalues():
                    subcomp._model = val
        elif name == '_parent':
            if val is not None and not isinstance(val, Block) and \
                    not (type(val) == weakref.ref and isinstance(val(), Block)):
                raise ValueError, "Cannot set the '_parent' attribute of a Block with name=%s to a non-Block object with type=%s; Did you introduce a model component named '_parent'?" % (self.name, type(val))
            self.__dict__['_parent'] = val
        elif name != 'components' and name != '_index' \
                 and isinstance(val, Component):
            #
            # If this is a component type, then simply set it
            #
            self._add_component(name,val)
        else:
            #
            # Try to set the value. This may fail if the attribute
            # does not already exist in this model, if the set_value
            # function is not defined, or if a bad value is
            # provided. In the latter case, a ValueError will be
            # thrown, which # we raise. Otherwise, this is an object
            # that we need to set directly.
            #
            if name in self.__dict__:
                if val is None:
                    del self.__dict__[name]
                    return
            
                setter = getattr(self.__dict__[name], 'set_value', None)
                if setter is None:
                    if isinstance(self.__dict__[name], Component):
                        raise Exception(
                            "Expected component %s to have a 'set_value' "
                            "method, but none was found." % \
                            ( self.__dict__[name], ))
                    self.__dict__[name]=val
                else:
                    setter(val)
            else:
                self.__dict__[name]=val
            #try:
            #    self.__dict__[name].set_value(val)
            #except ValueError, e:
            #    raise
            #except Exception, e:
            #    self.__dict__[name]=val

    def fix_all_vars(self):
        for var in self.components[Var]:
            self.components[Var][var].fix_all()
        for block in self.components[Block]:
            self.components[Block][block].fix_all_vars()

    def unfix_all_vars(self):
        for var in self.components[Var]:
            self.components[Var][var].unfix_all()
        for block in self.components[Block]:
            self.components[Block][block].unfix_all_vars()

    def pprint(self, filename=None, ostream=None, verbose=False):
        """
        Print a summary of the model info
        """
        self.components.pprint(filename=filename, ostream=ostream, verbose=verbose)

    def display(self, filename=None, ostream=None):
        """
        Print the Pyomo model in a verbose format.
        """
        if filename is not None:
            OUTPUT=open(filename,"w")
            self.display(ostream=OUTPUT)
            OUTPUT.close()
            return
        if ostream is None:
            ostream = sys.stdout
        if self._parent is not None and self._parent() is not None:
            print >>ostream, "Block "+self.name
        else:
            print >>ostream, "Model "+self.name
        #
        print >>ostream, ""
        print >>ostream, "  Variables:"
        VAR = self.active_components(Var)
        if len(VAR) == 0:
            print >>ostream, "    None"
        else:
            for ndx in VAR:
                VAR[ndx].display(prefix="    ",ostream=ostream)
        #
        print >>ostream, ""
        print >>ostream, "  Objectives:"
        OBJ = self.active_components(Objective)
        if len(OBJ) == 0:
            print >>ostream, "    None"
        else:
            for ndx in OBJ:
                OBJ[ndx].display(prefix="    ",ostream=ostream)
        print >>ostream, ""
        #
        CON = self.active_components(Constraint)
        print >>ostream, "  Constraints:"
        if len(CON) == 0:
            print >>ostream, "    None"
        else:
            for ndx in CON:
                CON[ndx].display(prefix="    ",ostream=ostream)



#
# Author:  Gabe Hackebeil
# Purpose: For regression testing to ensure that the Pyomo
#          NL writer properly modifies product expressions 
#          with only constant terms in the denominator (that
#          are involved in nonlinear expressions).
#          The ASL differentiation routines seem to have a 
#          bug that causes the lagrangian hessian to become
#          dense unless this constant term in moved to the 
#          numerator. 
#
#          This test model relies on the asl_test executable. It
#          will not solve if sent to a real optimizer.
#

from coopr.pyomo import *

model = ConcreteModel()
model.x = Var(bounds=(-1.0,1.0),initialize=1.0)
model.y = Var(bounds=(-1.0,1.0),initialize=2.0)
model.v = Var(bounds=(-1.0,1.0),initialize=3.0)
model.p = Var(initialize=2.0)
model.p.fixed = True

model.OBJ = Objective(expr=model.x)
model.CON1 = Constraint(expr=lambda model: (2.0,1.0/model.p*model.v*(model.x-model.y)))
model.CON2 = Constraint(expr=model.v*1.0/model.p*(model.x-model.y) == 2.0)
model.CON3 = Constraint(expr=model.v*(model.x-model.y)/model.p == 2.0)
model.CON4 = Constraint(expr=model.v*(model.x/model.p-model.y/model.p) == 2.0)
model.CON5 = Constraint(expr=model.v*(model.x-model.y)*(1.0/model.p) == 2.0)
model.CON6 = Constraint(expr=model.v*(model.x-model.y) == 2.0*model.p)



from coopr.pyomo import *

INDEX_SET1 = [1,2]
INDEX_SET2 = [1,2,'a']
PIECEWISE_PTS = dict([((t1,t2),[1.0,2.0,3.0]) for t1 in INDEX_SET1 for t2 in INDEX_SET2])
F = lambda x: x**2


def define_model():

    model = ConcreteModel()

    def SOS_indices_init(model,t1,t2):
        return [(t1,t2,i) for i in xrange(1,len(PIECEWISE_PTS[t1,t2]))]
    def ub_indices_init(model):
        return [(t1,t2,i) for t1 in INDEX_SET1 for t2 in INDEX_SET2 for i in xrange(1,len(PIECEWISE_PTS[t1,t2]))]
    
    model.SOS_indices = Set(INDEX_SET1,INDEX_SET2,dimen=3, ordered=True, initialize=SOS_indices_init)
    model.ub_indices = Set(ordered=True, dimen=3,initialize=ub_indices_init)

    model.x = Var(INDEX_SET1,INDEX_SET2)
    model.Fx = Var(INDEX_SET1,INDEX_SET2)
    #Add SOS1 variable to model
    model.y = Var(model.ub_indices,within=NonNegativeReals)

    def constraint1_rule(model,t1,t2,i):
        return model.Fx[t1,t2] - (F(PIECEWISE_PTS[t1,t2][i-1]) + ((F(PIECEWISE_PTS[t1,t2][i])-F(PIECEWISE_PTS[t1,t2][i-1]))/(PIECEWISE_PTS[t1,t2][i]-PIECEWISE_PTS[t1,t2][i-1]))*(model.x[t1,t2]-PIECEWISE_PTS[t1,t2][i-1])) <= 25.0*(1-model.y[t1,t2,i])
    def constraint2_rule(model,t1,t2):
        return sum(model.y[t1,t2,j] for j in xrange(1,len(PIECEWISE_PTS[t1,t2]))) == 1

    model.obj = Objective(expr=summation(model.Fx), sense=maximize)
    model.constraint1 = Constraint(model.ub_indices,rule=constraint1_rule)
    model.constraint2 = Constraint(INDEX_SET1,INDEX_SET2,rule=constraint2_rule)
    model.SOS_set_constraint = SOSConstraint(INDEX_SET1,INDEX_SET2, var=model.y, set=model.SOS_indices, sos=1)

    #Fix the answer for testing purposes
    model.set_answer_constraint1 = Constraint(expr= model.x[1,1] == 2.5)
    model.set_answer_constraint2 = Constraint(expr= model.x[2,1] == 1.5)
    model.set_answer_constraint3 = Constraint(expr= model.x[1,2] == 2.5)
    model.set_answer_constraint4 = Constraint(expr= model.x[2,2] == 1.5)
    model.set_answer_constraint5 = Constraint(expr= model.x[1,'a'] == 2.5)
    model.set_answer_constraint6 = Constraint(expr= model.x[2,'a'] == 1.5)

    return model

model = define_model()



#
# Imports
#

import sys
import os
from os.path import abspath, dirname
sys.path.insert(0, dirname(dirname(dirname(dirname(abspath(__file__))))))
from coopr.pyomo import *

infinity = float('inf')

#
# Model
#

model = AbstractModel()

model.FOOD = Set()

model.cost = Param(model.FOOD, within=PositiveReals)

model.f_min = Param(model.FOOD, within=NonNegativeReals, default=0.0)

MAX_FOOD_SUPPLY = 20.0 # McDonald's doesn't stock infinite food
def f_max_validate (model, value, j):
    return model.f_max[j] > model.f_min[j]
model.f_max = Param(model.FOOD, validate=f_max_validate, default=MAX_FOOD_SUPPLY)

# Unneeded vars - they're in the .dat file, so we list them here
model.NUTR = Set()
model.n_min = Param(model.NUTR, within=NonNegativeReals, default=0.0)
model.n_max = Param(model.NUTR, default=infinity)
model.amt = Param(model.NUTR, model.FOOD, within=NonNegativeReals)

# --------------------------------------------------------

def Buy_bounds(model, i):
    return (model.f_min[i], model.f_max[i])
model.Buy = Var(model.FOOD, bounds=Buy_bounds, within=NonNegativeIntegers)

# --------------------------------------------------------

def Total_Cost_rule(model):
    return sum(model.cost[j] * model.Buy[j] for j in model.FOOD)
model.Total_Cost = Objective(rule=Total_Cost_rule, sense=minimize)

# --------------------------------------------------------

def Entree_rule(model):
    entrees = ['Quarter Pounder w Cheese', 'McLean Deluxe w Cheese', 'Big Mac', 'Filet-O-Fish', 'McGrilled Chicken']
    return sum(model.Buy[e] for e in entrees) >= 1
model.Entree = Constraint(rule=Entree_rule)

def Side_rule(model):
    sides = ['Fries, small', 'Sausage McMuffin']
    return sum(model.Buy[s] for s in sides) >= 1
model.Side = Constraint(rule=Side_rule)

def Drink_rule(model):
    drinks = ['1% Lowfat Milk', 'Orange Juice']
    return sum(model.Buy[d] for d in drinks) >= 1
model.Drink = Constraint(rule=Drink_rule)



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

#
# AMPL Problem Writer Plugin
#

__all__ = ['ampl_var_id', 'intrinsic_function_expressions', 'ProblemWriter_nl', 'ampl_representation', 'generate_ampl_repn', 'generate_ef_ampl_repn', "_trivial_labeler"]


from coopr.opt import ProblemFormat, WriterFactory
from coopr.opt.base import *
from coopr.pyomo.base.objective import minimize, maximize
from coopr.pyomo.base import *
from coopr.pyomo.base import expr
from coopr.pyomo.base import intrinsic_functions
from coopr.pyomo.base.var import _VarData, Var, _VarElement
from coopr.pyomo.base.numvalue import NumericConstant, native_numeric_types
from coopr.pyomo.base.param import _ParamData
from coopr.pyomo.base import var
from coopr.pyomo.base import param
from pyutilib.component.core import alias
from pyutilib.misc import PauseGC

from coopr.opt.base import SymbolMap
import itertools

import time
import gc

import logging
logger = logging.getLogger('coopr.pyomo')

ampl_var_id = {}

intrinsic_function_expressions = {'log':'o43', 'log10':'o42', 'sin':'o41', 'cos':'o46', 'tan':'o38', 'sinh':'o40', 'cosh':'o45', 'tanh':'o37', 'asin':'o51', 'acos':'o53', 'atan':'o49', 'exp':'o44', 'sqrt':'o39', 'asinh':'o50', 'acosh':'o52', 'atanh':'o47', 'pow':'o5'}

class _trivial_labeler(object):
    def __call__(self, obj, *args):
        label = str(obj.name)
        try:
            block = obj.component
        except:
            block = None
        
        if block is None:
            return nl_label_from_name(label)
        # At this point we have confirmed that block is a component
        block = block()._parent
        if block is None:
            return nl_label_from_name(label)
        # At this point we have confirmed that the component belongs
        # to something, like a model or another block

        # GAH: If this model is flat, i.e., no nested blocks
        #      then don't generate label names with prepended
        #      parent names
        if block()._parent is None:
            return nl_label_from_name(label)

#        block = block()
        while block is not None:
            block = block()
            label = block.name + "." + label
            block = block._parent
        return nl_label_from_name(label)

class ModelSOS(object):
    def __init__(self):
        self.sosno_list = []
        self.ref_list = []
        self.block_cntr = 0

    def count_constraint(self,con,symbol_map, labeler,level,index=None):

        # Identifies the the current set from others in the NL file
        self.block_cntr += 1

        # If SOS1, the identifier must be positive
        # if SOS2, the identifier must be negative
        sign_tag = 1
        if level == 2:
            sign_tag = -1

        # The name of the variable being indexed
        prefix = symbol_map.getSymbol(con.sos_vars()._model(), labeler)
        varName = prefix + "_" + str(con.sos_vars())

        if index is None:
            tmpSet = con.sos_set()
        else:
            tmpSet = con.sos_set()[index]

        # Get all the variables
        cntr = 0
        for x in tmpSet:
            ID = ampl_var_id[symbol_map.getSymbol( con.sos_vars()[x], labeler )]
            self.sosno_list.append((ID,self.block_cntr*sign_tag))
            # We need to weight each variable
            # For now we just increment a counter
            if cntr > 0:
                # We only need to report variables with weight greater than zero
                # Zero weight is assumed for variables with no weight reported.
                self.ref_list.append((ID,cntr))
            cntr += 1


class ProblemWriter_nl(AbstractProblemWriter):

    alias(str(ProblemFormat.nl))

    def __init__(self):
        AbstractProblemWriter.__init__(self, ProblemFormat.nl)

    def __call__(self, model, filename, solver_capability, generate_symbolic_labels, **kwds):
        # Pause the GC for the duration of this method
        suspend_gc = PauseGC()
        if filename is None:
            filename = model.name + ".nl"
        OUTPUT=open(filename,"w")
        ampl_var_id.clear()
        symbol_map = self._print_model_NL(model, OUTPUT, solver_capability, generate_symbolic_labels, verbose=False, gen_obj_ampl_repn=True, gen_con_ampl_repn=True)
        OUTPUT.close()
        return filename, symbol_map

    def _get_bound(self, exp):
        if exp.is_fixed():
            return exp()
        else:
            raise ValueError, "ERROR: nonconstant bound: " + str(exp)
            return None

        exp_type = type(exp)
        # Params & Constants are fixed, so are handled above
        #if exp_type is NumericConstant:
        #    return exp.value
        #elif exp_type is _ParamData:
        #    return exp.value
        if exp_type is expr._SumExpression:
            sum_expr_value=0.0
            for i in xrange(len(exp._args)):
                sum_expr_value += self._get_bound(exp._args[i])*exp._coef[i]
            return sum_expr_value
        elif exp_type is expr._ProductExpression:
            prod_expr_value=1.0
            for i in xrange(len(exp._numerator)):
                prod_expr_value *= float(self._get_bound(exp._numerator[i]))
            for i in xrange(len(exp._denominator)):
                prod_expr_value /= float(self._get_bound(exp._denominator[i]))
            return prod_expr_value
        else:
            raise ValueError, "ERROR: nonconstant bound: " + str(exp)
            return None

    def _print_nonlinear_terms_NL(self, OUTPUT, exp, symbol_map, labeler):

        if isinstance(exp, expr._IntrinsicFunctionExpression):
            if exp.name in intrinsic_function_expressions:
                print >>OUTPUT, intrinsic_function_expressions[exp.name]
            else:
                logger.error("Unsupported intrinsic function ({0})", exp.name)
                raise TypeError("ASL writer does not support '{0}' expressions"
                               .format(exp.name))
                
            for child_exp in exp._args:
                self._print_nonlinear_terms_NL(OUTPUT, child_exp, symbol_map, labeler)

        elif isinstance(exp, expr._SumExpression):
            n = len(exp._args)

            if n > 2:
#                print >>OUTPUT, "o54  #sum"
                print >>OUTPUT, "o54"
                print >>OUTPUT, n

                if exp._const != 0.0:
#                    print >>OUTPUT, "o0  #+"
                    print >>OUTPUT, "o0"
                    print >>OUTPUT, "n{0}".format(exp._const)

                for i in xrange(0,n):
                    if exp._coef[i] != 1:
#                        print >>OUTPUT, "o2  #*"
                        print >>OUTPUT, "o2"
                        print >>OUTPUT, "n{0}".format(exp._coef[i])
                    self._print_nonlinear_terms_NL(OUTPUT, exp._args[i], symbol_map, labeler)
            else:
                if exp._const != 0.0:
#                    print >>OUTPUT, "o0  #+"
                    print >>OUTPUT, "o0"
                    print >>OUTPUT, "n{0}".format(exp._const)

                for i in xrange(0,n-1):
#                    print >>OUTPUT, "o0  #+"
                    print >>OUTPUT, "o0"
                    if exp._coef[i] != 1:
#                        print >>OUTPUT, "o2  #*"
                        print >>OUTPUT, "o2"
                        print >>OUTPUT, "n{0}".format(exp._coef[i])
                    self._print_nonlinear_terms_NL(OUTPUT, exp._args[i], symbol_map, labeler)

                if exp._coef[n-1] != 1:
#                    print >>OUTPUT, "o2  #*"
                    print >>OUTPUT, "o2"
                    print >>OUTPUT, "n{0}".format(exp._coef[n-1])
                self._print_nonlinear_terms_NL(OUTPUT, exp._args[n-1], symbol_map, labeler)

        elif isinstance(exp, list):
            # this is an implied summation of expressions (we did not create a new sum expression for efficiency)
            # this should be a list of tuples where [0] is the coeff and [1] is the expr to write
            n = len(exp)
            if n > 2:
#                print >>OUTPUT, "o54  #sum"
                print >>OUTPUT, "o54"
                print >>OUTPUT, n

                for i in xrange(0,n):
                    assert(isinstance(exp[i],tuple))
                    coef = exp[i][0]
                    child_exp = exp[i][1]

                    if coef != 1.0:
#                        print >>OUTPUT, "o2  #*"
                        print >>OUTPUT, "o2"
                        print >>OUTPUT, "n{0}".format(coef)
                    self._print_nonlinear_terms_NL(OUTPUT, child_exp, symbol_map, labeler)
            else:
                for i in xrange(0,n):
                    assert(isinstance(exp[i],tuple))
                    coef = exp[i][0]
                    child_exp = exp[i][1]

                    if i != n-1:
                        # need the + op if it is not the last entry in the list
#                        print >>OUTPUT, "o0  #+"
                        print >>OUTPUT, "o0"

                    if coef != 1.0:
#                        print >>OUTPUT, "o2  #*"
                        print >>OUTPUT, "o2"
                        print >>OUTPUT, "n{0}".format(coef)
                    self._print_nonlinear_terms_NL(OUTPUT, child_exp, symbol_map, labeler)

        elif isinstance(exp, expr._ProductExpression):
            denom_exists = False
            if len(exp._denominator) == 0:
                pass
            else:
#                print >>OUTPUT, "o3 # /"
                print >>OUTPUT, "o3"
                denom_exists = True

            if exp.coef != 1:
#                print >>OUTPUT, "o2 # *"
                print >>OUTPUT, "o2"
                print >>OUTPUT, "n{0}".format(exp.coef)

            if len(exp._numerator) == 0:
                print >>OUTPUT, "n1"

            # print out the numerator
            child_counter = 0
            max_count = len(exp._numerator)-1
            for child_exp in exp._numerator:
                if child_counter < max_count:
#                    print >>OUTPUT, "o2  #*"
                    print >>OUTPUT, "o2"
                self._print_nonlinear_terms_NL(OUTPUT, child_exp, symbol_map, labeler)
                child_counter += 1

            if denom_exists:
            # print out the denominator
                child_counter = 0
                max_count = len(exp._denominator)-1
                for child_exp in exp._denominator:
                    if child_counter < max_count:
#                        print >>OUTPUT, "o2  #*"
                        print >>OUTPUT, "o2"
                    self._print_nonlinear_terms_NL(OUTPUT, child_exp, symbol_map, labeler)
                    child_counter += 1

        elif isinstance(exp, expr._PowExpression):
#            print >>OUTPUT, "o5  #^"
            print >>OUTPUT, "o5"
            for child_exp in exp._args:
                self._print_nonlinear_terms_NL(OUTPUT, child_exp, symbol_map, labeler)

        elif (isinstance(exp,var._VarData) or isinstance(exp,var.Var)) and not exp.is_fixed():
#            print >>OUTPUT, "v{0} #{1}".format(ampl_var_id[exp.name], exp.name)
#            print >>OUTPUT, "v{0} #{1}".format(ampl_var_id[exp.name], exp.name)
            variable_label = symbol_map.getSymbol(exp, labeler)
            print >>OUTPUT, "v{0}".format(ampl_var_id[variable_label])

        elif isinstance(exp,param._ParamData):
#            print >>OUTPUT, "n{0} #{1}".format(exp.value, exp.name)
            print >>OUTPUT, "n{0}".format(value(exp))

        elif (isinstance(exp,NumericConstant) or exp.is_fixed):
#            print >>OUTPUT, "n{0} # numeric constant".format(exp.value)
            print >>OUTPUT, "n{0}".format(value(exp))

        else:
            raise ValueError, "Unsupported expression type in _print_nonlinear_terms_NL"
        
    
    def _print_model_NL(self, model, OUTPUT, solver_capability, generate_symbolic_labels, verbose=False, gen_obj_ampl_repn=True, gen_con_ampl_repn=True):

        # maps NL variables to the "real" variable names in the problem.
        # it's really NL variable ordering, as there are no variable names
        # in the NL format. however, we by convention make them go from
        # x0 upward.

#        start_overall_time = time.clock()

#        if type(model) is dict:
#            symbol_map = self._print_ef(model, OUTPUT)
#            return symbol_map

        # because NL preprocessing occurs during the write process, we don't have direct access to
        # this method through the solve() method of a solver plugin. and keyword propagation is too
        # complex to think that would be a good solution. so, we can optionally tag the input
        # instance with the preprocessing keywords.
        if hasattr(model, "gen_obj_ampl_repn"):
            gen_obj_ampl_repn = getattr(model, "gen_obj_ampl_repn")
            delattr(model, "gen_obj_ampl_repn")
        if hasattr(model, "gen_con_ampl_repn"):
            gen_con_ampl_repn = getattr(model, "gen_con_ampl_repn")
            delattr(model, "gen_con_ampl_repn")            
            
        symbol_map = SymbolMap(model)
        labeler = _trivial_labeler()

        #
        # Collect statistics
        #
        Obj = model.active_components(Objective)

        Vars = dict() # will be the entire list of all vars used in the problem
        # linear variables
        LinearVars = set()
        LinearVarsInt = set()
        LinearVarsBool = set()
        #
        # Count number of objectives and build the ampl_repns
        #

#        print "Generate objective representation:",
#        start_time = time.clock()

        n_objs = 0
        n_nonlinear_objs = 0
        ObjVars = set()
        ObjNonlinearVars = set()
        ObjNonlinearVarsInt = set()
        for obj in Obj:
            for i in Obj[obj]:
                tmp_obj = Obj[obj][i]

                if gen_obj_ampl_repn is True:
                    # See note in loop for constraints as to why the expression is cloned here
                    if isinstance(tmp_obj.expr,expr.Expression):
                        ampl_repn = generate_ampl_repn(tmp_obj.expr.clone(), symbol_map, labeler)
                    else:
                        # No clone is necessary, this is either a variable or a constant.
                        ampl_repn = generate_ampl_repn(tmp_obj.expr, symbol_map, labeler)
                    tmp_obj.ampl_repn = ampl_repn
                else:
                    ampl_repn = tmp_obj.ampl_repn
                Vars.update(ampl_repn._linear_terms_var)
                Vars.update(ampl_repn._nonlinear_vars)

                LinearVars.update(ampl_repn._linear_terms_var)
                ObjNonlinearVars.update(ampl_repn._nonlinear_vars)
                    
                ObjVars.update(ampl_repn._linear_terms_var)
                ObjVars.update(ampl_repn._nonlinear_vars)
                n_objs += 1
                if ampl_repn.is_nonlinear():
                    n_nonlinear_objs += 1

        if n_objs > 1:
            raise ValueError, "asl solver has detected multiple objective functions, but currently only handles a single objective."

#        end_time = time.clock()
#        print (end_time - start_time)

#        print "Generate constraint repns:",
#        start_time = time.clock()

        #
        # Count number of constraints and build the ampl_repns
        #

        n_ranges = 0
        n_single_sided_ineq = 0
        n_equals = 0
        n_nonlinear_constraints = 0
        ConNonlinearVars = set()
        ConNonlinearVarsInt = set()
        nnz_grad_constraints = 0
        constraint_bounds_dict = {}
        stack_lin = {}
        stack_nonlin = {}
        nonlin_con_order_list = []
        lin_con_order_list = []

#        Con = model.active_components(Constraint)
#        Con.update( model.active_components(ConstraintList) )
        
        for block in model.all_blocks():
            active_constraints = block.active_components(Constraint)
            active_constraintlists = block.active_components(ConstraintList)
            for constraint in itertools.chain(active_constraints.itervalues(), active_constraintlists.itervalues()):
                ### GAH: Removing this check so that the NL writer does not depend whether the canonical repn
                ###      was generated. If we are going to write out trivial constraints as a feasibility check
                ###      for fixed variables then we must not skip the next inner loop.
                ###      Trivial constraints can be identified using the ampl_repn.
                ### ****NOTE: If the canonical repn is going to be generated we could probably optimize
                ###           the inner part of the next loop to manually extract what we need from the 
                ###           canonical repn to quickly build the trivial ampl_repn, but this is not currently
                ###           done.
                #if constraint.trivial is True:
                #    continue
                
                for index in sorted(constraint.keys()):
                    constraint_data = constraint[index]
                    
                    if not constraint_data.active:
                        continue

                    con_name = symbol_map.getSymbol( constraint_data, labeler )
                    if gen_con_ampl_repn is True:
                        ampl_repn = generate_ampl_repn(constraint_data.body, symbol_map, labeler)
                        constraint_data.ampl_repn = ampl_repn
                    else:
                        ampl_repn = constraint_data.ampl_repn

                    linear_vars = ampl_repn._linear_terms_var
                    nonlinear_vars = ampl_repn._nonlinear_vars
                   
 
                    ### GAH: Even if this is fixed, it is still usefule to write out these types of constraints
                    ###      (trivial) as a feasibility check for fixed variables, in which case the solver
                    ###      will pick up on the model infeasibility. Therefore, we will fix the code in the 
                    ###      "J" section to handle the case where no variables exist in the ampl_repn 
                    ###      (i.e ampl_repn.is_fixed() is True).
                    #if ampl_repn.is_fixed():
                    #    continue
                    
                    if ampl_repn.is_nonlinear():
                        stack_nonlin[con_name] = constraint_data
                        nonlin_con_order_list.append(con_name)
                        n_nonlinear_constraints += 1
                    else:
                        stack_lin[con_name] = constraint_data
                        lin_con_order_list.append(con_name)
    
                    # Merge variables from this expression
                    all_vars = {}
                    all_vars.update(linear_vars)
                    all_vars.update(nonlinear_vars)
                    Vars.update(all_vars)
                    nnz_grad_constraints += len(all_vars)
    
                    LinearVars.update(linear_vars)
                    ConNonlinearVars.update(nonlinear_vars)
    
                    L = None
                    U = None
                    if constraint_data.lower is not None: L = self._get_bound(constraint_data.lower)
                    if constraint_data.upper is not None: U = self._get_bound(constraint_data.upper)
                    offset = ampl_repn._constant
                    if L == U:
                        if (L is None):
#                            constraint_bounds_dict[con_name] = [3, '', '', con_name]
                            constraint_bounds_dict[con_name] = [3, '', '']
                            # constraint with NO bounds ???
                            raise ValueError, "Constraint " + str(con) +\
                                             "[" + str(i) + "]: lower bound = None and upper bound = None"
                        else:
#                            constraint_bounds_dict[con_name] = [4, L-offset, '', con_name]
                            constraint_bounds_dict[con_name] = [4, L-offset, '']
                            n_equals += 1
                    elif L is None:
#                        constraint_bounds_dict[con_name] = [1, '', U-offset, con_name]
                        constraint_bounds_dict[con_name] = [1, '', U-offset]
                        n_single_sided_ineq += 1
                    elif U is None:
#                        constraint_bounds_dict[con_name] = [2, L-offset, '', con_name]
                        constraint_bounds_dict[con_name] = [2, L-offset, '']
                        n_single_sided_ineq += 1
                    elif (L > U):
                        msg = 'Constraint {0}: lower bound greater than upper' \
                              ' bound ({1} > {2})'
                        raise ValueError, msg.format(con_name, str(L), str(U))
                    else:
#                        constraint_bounds_dict[con_name] = [0, L-offset, U-offset, con_name]
                        constraint_bounds_dict[con_name] = [0, L-offset, U-offset]
                        # double sided inequality
                        # both are not none and they are valid
                        n_ranges += 1


        LinearVars = LinearVars.difference(ObjNonlinearVars)
        LinearVars = LinearVars.difference(ConNonlinearVars)

        if model.statistics.number_of_binary_variables != 0 or model.statistics.number_of_integer_variables != 0:
            var_names = LinearVars.copy()
            for var_name in var_names:
                var = Vars[var_name]
                var_type = type(var.domain)
                if var_type is IntegerSet:
                    LinearVarsInt.add(var_name)
                    LinearVars.discard(var_name)
                elif var_type is BooleanSet:
                    L = var.lb
                    U = var.ub
                    if L is None or U is None:
                        raise ValueError, "Variable " + str(var.name) +\
                            "is binary, but does not have lb and ub set"
                    elif value(L) == 0 and value(U) == 1:
                        LinearVarsBool.add(var_name)
                        LinearVars.discard(var_name)
                    else:
                        LinearVarsInt.add(var_name)
                        LinearVars.discard(var_name)

            var_names = ObjNonlinearVars.copy()
            for var_name in var_names:
                var_type = type(Vars[var_name].domain)
                if var_type is IntegerSet or var_type is BooleanSet:
                    ObjNonlinearVarsInt.add(var_name)
                    ObjNonlinearVars.discard(var_name)
            var_names = ConNonlinearVars.copy()
            for var_name in var_names:
                var_type = type(Vars[var_name].domain)
                if var_type is IntegerSet or var_type is BooleanSet:
                    ConNonlinearVarsInt.add(var_name)
                    ConNonlinearVars.discard(var_name)

        Nonlinear_Vars_in_Objs_and_Constraints = ObjNonlinearVars.intersection(ConNonlinearVars)
        Discrete_Nonlinear_Vars_in_Objs_and_Constraints = ObjNonlinearVarsInt.intersection(ConNonlinearVarsInt)
        ObjNonlinearVars = ObjNonlinearVars.difference(Nonlinear_Vars_in_Objs_and_Constraints)
        ConNonlinearVars = ConNonlinearVars.difference(Nonlinear_Vars_in_Objs_and_Constraints)
        ObjNonlinearVarsInt = ObjNonlinearVarsInt.difference(Discrete_Nonlinear_Vars_in_Objs_and_Constraints)
        ConNonlinearVarsInt = ConNonlinearVarsInt.difference(Discrete_Nonlinear_Vars_in_Objs_and_Constraints)
        

        # put the ampl variable id into the variable
        var_id_ctr=0
        full_var_list = []
        full_var_list.extend(sorted(Nonlinear_Vars_in_Objs_and_Constraints))
        full_var_list.extend(sorted(Discrete_Nonlinear_Vars_in_Objs_and_Constraints))
        idx_nl_both = len(full_var_list)
        full_var_list.extend(sorted(ConNonlinearVars))
        full_var_list.extend(sorted(ConNonlinearVarsInt))
        idx_nl_con = len(full_var_list)
        full_var_list.extend(sorted(ObjNonlinearVars))
        full_var_list.extend(sorted(ObjNonlinearVarsInt))
        idx_nl_obj = len(full_var_list)
        full_var_list.extend(sorted(LinearVars))
        full_var_list.extend(sorted(LinearVarsBool))
        full_var_list.extend(sorted(LinearVarsInt))

        if (idx_nl_obj == idx_nl_con):
            idx_nl_obj = idx_nl_both

        for var_name in full_var_list:
            ampl_var_id[var_name] = var_id_ctr
            var_id_ctr += 1

#        end_time = time.clock()
#        print (end_time - start_time)

        colfilename = None
        if OUTPUT.name.endswith('.nl'):
            colfilename = OUTPUT.name.replace('.nl','.col')
        else:
            colfilename = OUTPUT.name+'.col'
        if generate_symbolic_labels is True:
            colf = open(colfilename,'w')
            for var_name in full_var_list:
                print >>colf, var_name
            colf.close()
                
#        print "Printing NL file header:",
#        start_time = time.clock()
        #
        # Print Header
        #
        # LINE 1
        #
        print >>OUTPUT,"g3 1 1 0\t# problem {0}".format(model.name)
        #
        # LINE 2
        #
        print >>OUTPUT, " {0} {1} {2} {3} {4} \t# vars, constraints, objectives, general inequalities, equalities  " \
           .format(len(Vars), n_single_sided_ineq+n_ranges+n_equals, n_objs, n_ranges, n_equals)
        #
        # LINE 3
        #
        print >>OUTPUT, " {0} {1} \t# nonlinear constraints, objectives".format(n_nonlinear_constraints, n_nonlinear_objs)
        #
        # LINE 4
        #
        print >>OUTPUT, " 0 0\t# network constraints: nonlinear, linear"
        #
        # LINE 5
        #
        nlv_both = len(Discrete_Nonlinear_Vars_in_Objs_and_Constraints)+len(Nonlinear_Vars_in_Objs_and_Constraints)
        print >>OUTPUT, " {0} {1} {2} \t# nonlinear vars in constraints, objectives, both" \
        .format(idx_nl_con, idx_nl_obj, idx_nl_both)

        #
        # LINE 6
        #
        print >>OUTPUT, " 0 0 0 1\t# linear network variables; functions; arith, flags"
        #
        # LINE 7
        #
        n_int_nonlinear_b = len(Discrete_Nonlinear_Vars_in_Objs_and_Constraints)
        n_int_nonlinear_c = len(ConNonlinearVarsInt)
        n_int_nonlinear_o = len(ObjNonlinearVarsInt) 
        print >>OUTPUT, " {0} {1} {2} {3} {4} \t# discrete variables: binary, integer, nonlinear (b,c,o)" \
           .format(len(LinearVarsBool), len(LinearVarsInt), n_int_nonlinear_b, n_int_nonlinear_c, n_int_nonlinear_o)
        #
        # LINE 8
        #
        # objective info computed above
        print >>OUTPUT, " {0} {1} \t# nonzeros in Jacobian, obj. gradient".format(nnz_grad_constraints, len(ObjVars))
        #
        # LINE 9
        #
        print >>OUTPUT, " 0 0\t# max name lengths: constraints, variables"
        #
        # LINE 10
        #
        print >>OUTPUT, " 0 0 0 0 0\t# common exprs: b,c,o,c1,o1"

#        end_time = time.clock()
#        print (end_time - start_time)

#        print "Printing constraints:",
#        start_time = time.clock()


        # SOS constraints - largely taken from cpxlp.py so updates there,
        # should be applied here
        # TODO: Allow users to specify the variables coefficients for custom
        # branching/set orders - refer to cpxlp.py
        sosn = solver_capability("sos1")
        sos1 = solver_capability("sos2")
        sos2 = solver_capability("sosn")

        if sosn or sos1 or sos2:
            modelSOS = ModelSOS()
            for block in model.all_blocks():
                constrs = block.active_components(SOSConstraint)
                for name, con in constrs.iteritems():
                    level = con.sos_level()
                    if (level == 1 and sos1) or (level == 2 and sos2) or (sosn):
                        masterIndex = con.sos_set_set()
                        if None in masterIndex:
                            modelSOS.count_constraint(con, symbol_map, labeler, level)
                        else:
                            for index in masterIndex:
                                modelSOS.count_constraint(con, symbol_map, labeler, level, index)
                                
            if len(modelSOS.sosno_list) > 0:
                print >>OUTPUT, 'S0', len(modelSOS.sosno_list), 'sosno'
                for sos_var_id, sos_var_block in modelSOS.sosno_list:
                    print >>OUTPUT, sos_var_id, sos_var_block
                print >>OUTPUT, 'S0', len(modelSOS.ref_list), 'ref'
                for sos_var_id, sos_var_weight in modelSOS.ref_list:
                    print >>OUTPUT, sos_var_id, sos_var_weight
            
        #
        # "C" lines
        #
        nc = 0

        cu = [0 for i in xrange(len(Vars))]
        
        rowfilename = None
        if OUTPUT.name.endswith('.nl'):
            rowfilename = OUTPUT.name.replace('.nl','.row')
        else:
            rowfilename = OUTPUT.name+'.row'
        if generate_symbolic_labels is True:
            rowf = open(rowfilename,'w')

        constraint_bounds_list = []
        for con_name in nonlin_con_order_list:
            ampl_repn = stack_nonlin[con_name].ampl_repn
            #### GAH: An ampl_repn that is fixed would have never been classified
            ####      as nonlinear in the first place, so this check is unnecessary
            #if ampl_repn.is_fixed():
            #    continue
            #else:
            print >>OUTPUT, "C{0}\t#{1}".format(nc, con_name)
            if generate_symbolic_labels is True:
                print >>rowf, con_name 
            cbd = constraint_bounds_dict[con_name]
#           bounds_string = "{0} {1} {2}  # c{3} {4}".format(cbd[0], cbd[1], cbd[2], nc, cbd[3])
            bounds_string = "{0} {1} {2}".format(cbd[0], cbd[1], cbd[2])
            constraint_bounds_list.append(bounds_string)
            #symbol_map["c" + str(nc)] = constraint_data.label
            symbol_map.alias(stack_nonlin[con_name], "c"+str(nc))
            nc += 1
            
            self._print_nonlinear_terms_NL(OUTPUT, ampl_repn._nonlinear_expr, symbol_map, labeler)
            ####

            con_vars = set(ampl_repn._linear_terms_var)
            con_vars.update(ampl_repn._nonlinear_vars)
            for d in con_vars:
                cu[ampl_var_id[d]] += 1

        for con_name in lin_con_order_list:
            ampl_repn = stack_lin[con_name].ampl_repn
            con_vars = set(ampl_repn._linear_terms_var)
            for d in con_vars:
                cu[ampl_var_id[d]] += 1
#            print >>OUTPUT, "C{0}\t#{1}[{2}]".format(nc,key,ndx)
            print >>OUTPUT, "C{0}".format(nc)
            if generate_symbolic_labels is True:
                print >>rowf, con_name
            cbd = constraint_bounds_dict[con_name]
#            bounds_string = "{0} {1} {2}  # c{3} {4}".format(cbd[0], cbd[1], cbd[2], nc, cbd[3])
            bounds_string = "{0} {1} {2}".format(cbd[0], cbd[1], cbd[2])
            constraint_bounds_list.append(bounds_string)
            #symbol_map["c" + str(nc)] = C[ndx].label
            symbol_map.alias(stack_lin[con_name], "c"+str(nc))
            nc += 1
            print >>OUTPUT, "n0"

        if generate_symbolic_labels is True:
            rowf.close()

#        end_time = time.clock()
#        print (end_time - start_time)

#        print "Printing objective:",
#        start_time = time.clock()
        #
        # "O" lines
        #
        no = 0
        for key in Obj:
            k = 0
            if Obj[key].sense == maximize:
                k = 1
            for ndx in Obj[key]:
                ampl_repn = Obj[key][ndx].ampl_repn
                if ampl_repn is None:
                    continue
#                print >>OUTPUT, "O{0} {1}\t#{2}[{3}]".format(no, k, key, ndx)
                print >>OUTPUT, "O{0} {1}".format(no, k)
                no += 1

                if ampl_repn.is_linear():
                    print >>OUTPUT, "n{0}".format(ampl_repn._constant)
                else:
                    if ampl_repn._constant != 0.0:
#                        print >>OUTPUT, "o0  #+"
                        print >>OUTPUT, "o0"  #+"
                        print >>OUTPUT, "n{0}".format(ampl_repn._constant)

                    self._print_nonlinear_terms_NL(OUTPUT, ampl_repn._nonlinear_expr, symbol_map, labeler)

#        end_time = time.clock()
#        print (end_time - start_time)

#        print "Printing intializations:",
#        start_time = time.clock()
        #
        # "x" lines
        #
        # variable initialization
        tmp_map = {}
        var_bound_list = []
        for var_name in full_var_list:
            var = Vars[var_name]
            # FIXME: This will need to change when var.value changes/goes away
            #        The test is here because we are inconsistent about initializing
            #        values of variables
            if var.value is not None:
#                tmp_map[ampl_var_id[var_name]] = [var.name,var.value]
                tmp_map[ampl_var_id[var_name]] = var.value
            if isinstance(var.domain, BooleanSet):
#                tmp_string = "0 0 1  # v{0}  {1}".format(ampl_var_id[var_name], var_name)
                tmp_string = "0 0 1"
                var_bound_list.append(tmp_string)
                #symbol_map["v"+str(ampl_var_id[var_name])] = var.label
                symbol_map.getSymbol(var, labeler, "v"+str(ampl_var_id[var_name]))
                symbol_map.alias(var, "v"+str(ampl_var_id[var_name]))
                continue
            L = var.lb
            U = var.ub
            if L is not None:
                Lv = value(L)
                if U is not None:
                    Uv = value(U)
                    if Lv == Uv:
#                        tmp_string = "4 {0}  # v{1}  {2}".format(Lv, ampl_var_id[var_name], var.name)
                        tmp_string = "4 {0}".format(Lv)
                        var_bound_list.append(tmp_string)
                    else:
#                        tmp_string = "0 {0} {1}  # v{2}  {3}".format(Lv, Uv, ampl_var_id[var_name], var.name)
                        tmp_string = "0 {0} {1}".format(Lv, Uv)
                        var_bound_list.append(tmp_string)
                else:
#                    tmp_string = "2 {0}  # v{1}  {2}".format(Lv, ampl_var_id[var_name], var.name)
                    tmp_string = "2 {0}".format(Lv)
                    var_bound_list.append(tmp_string)
            elif U is not None:
#                tmp_string = "1 {0}  # v{1}  {2}".format(value(U), ampl_var_id[var_name], var.name)
                tmp_string = "1 {0}".format(value(U))
                var_bound_list.append(tmp_string)
            else:
#                tmp_string = "3  #v{0}  {1}".format(ampl_var_id[var_name], var.name)
                tmp_string = "3"
                var_bound_list.append(tmp_string)
            #symbol_map["v"+str(ampl_var_id[var_name])] = var.label
            symbol_map.getSymbol(var, labeler, "v"+str(ampl_var_id[var_name]))
            symbol_map.alias(var, "v"+str(ampl_var_id[var_name]))


        print >>OUTPUT, "x{0}".format(len(tmp_map))
        for i in sorted(tmp_map.keys()):
#            print >>OUTPUT, "{0} {1} # {2} initial".format(i,tmp_map[i][1], tmp_map[i][0])
            print >>OUTPUT, i, tmp_map[i]
        
#        end_time = time.clock()
#        print (end_time - start_time)

#        print "Printing constraint bounds:",
#        start_time = time.clock()
        #
        # "r" lines
        #
        print >>OUTPUT, "r"
        for statement in constraint_bounds_list:
            print >>OUTPUT, statement

#        end_time = time.clock()
#        print (end_time - start_time)

#        print "Printing variable bounds:",
#        start_time = time.clock()
        #
        # "b" lines
        #
        print >>OUTPUT, "b"
        for bound in var_bound_list:
            print >>OUTPUT, bound

#        end_time = time.clock()
#        print (end_time - start_time)

#        print "Printing k's:",
#        start_time = time.clock()        

        #
        # "k" lines
        #
        ktot = 0
        n1 = len(Vars) - 1
        print >>OUTPUT, "k{0}".format(n1)
        ktot = 0
        for i in xrange(n1):
            ktot += cu[i]
            print >>OUTPUT, ktot

#        end_time = time.clock()
#        print (end_time - start_time)

#        print "Printing Jacobian's:",
#        start_time= time.clock()

        #
        # "J" lines
        # 
        nc = 0
        stack_ = stack_nonlin
        stack_.update(stack_lin)
        con_order_list = nonlin_con_order_list
        con_order_list.extend(lin_con_order_list)
        for con_name in con_order_list:
            ampl_repn = stack_[con_name].ampl_repn
            con_vars = set(ampl_repn._linear_terms_var)
            con_vars.update(ampl_repn._nonlinear_vars)
            ### GAH: This replaces the checks for constraint.trivial,
            ###      and ampl_repn.fixed. In these cases no active variables
            ###      will exist inside the ampl_repn. The purpose here is to have
            ###      constraints that become trivial due to fixed variables still
            ###      be written in the "C" section as a feasibility check for the fixed
            ###      variable values.
            if len(con_vars) > 0:
#                print >>OUTPUT, "J{0} {1} # {2}".format(nc, len(con_vars), con_name)
                print >>OUTPUT, "J{0} {1}".format(nc, len(con_vars))
                nc += 1
            
                for con_var in sorted(con_vars):
                    if con_var in ampl_repn._linear_terms_var:
                        print >>OUTPUT, ampl_var_id[con_var], ampl_repn._linear_terms_coef[con_var]
                    else:
                        print >>OUTPUT, ampl_var_id[con_var], 0

#        end_time = time.clock()
#        print (end_time - start_time)

#        print "Printing objective gradient:",
#        start_time = time.clock()
        #
        # "G" lines
        #
        no = 0
        grad_entries = {}
        for key in Obj:
            if Obj[key].trivial:
                continue
            obj = Obj[key]
            for ndx in obj:
                ampl_repn = obj[ndx].ampl_repn
                obj_vars = set(ampl_repn._linear_terms_var)
                obj_vars.update(ampl_repn._nonlinear_vars)
                if len(obj_vars) == 0:
                    pass
                else:
                    print >>OUTPUT, "G{0} {1}".format(no, len(obj_vars))
                no += 1

                for obj_var in obj_vars:
                    if obj_var in ampl_repn._linear_terms_var:
                        grad_entries[ampl_var_id[obj_var]] = ampl_repn._linear_terms_coef[obj_var]
                    else:
                        grad_entries[ampl_var_id[obj_var]] = 0

        for var_id in sorted(grad_entries.keys()):
            print >>OUTPUT, var_id, grad_entries[var_id]

#        print (end_time - start_time)
#        end_overall_time = time.clock()
#        print "Total time:",end_overall_time-start_overall_time

        return symbol_map


class ampl_representation:
    __slots__ = ('_constant', '_linear_terms_var', '_linear_terms_coef', '_nonlinear_expr', '_nonlinear_vars', '_name_prefix')
    def __init__(self):
        self._constant = 0
        self._linear_terms_var = {}
        self._linear_terms_coef = {}
        self._nonlinear_expr = None
        self._nonlinear_vars = {}
        self._name_prefix = ""

    def clone(self):
        clone = ampl_representation()
        clone._constant = self._constant
        clone._linear_terms_var = self._linear_terms_var
        clone._linear_terms_coef = self._linear_terms_coef
        clone._nonlinear_expr = self._nonlinear_expr
        clone._nonlinear_vars = self._nonlinear_vars
        return clone

    # for debugging
    def __str__(self):
        import StringIO
        output = StringIO.StringIO()
        output.write("\n")
        output.write("constant:       "+str(self._constant)+"\n")
        output.write("linear vars:    "+str(self._linear_terms_var)+"\n")
        output.write("linear coef:    "+str(self._linear_terms_coef)+"\n")
        output.write("nonlinear expr:\n")
        try:
            self._nonlinear_expr.pprint(ostream=output)
        except AttributeError:
            output.write(str(self._nonlinear_expr)+"\n")
        output.write("nonlinear vars: "+str(self._nonlinear_vars)+"\n")
        output.write("\n")
        ret_str = output.getvalue()
        output.close()
        return ret_str

    def __eq__(self, other):
        # Can only be equal to other ampl_representation instances
        if not isinstance(other, ampl_representation):
            return False
        
        # Immediately check constants
        if self._constant != other._constant:
            return False

        # Optimization: check linear term lengths before iterating
        if len(self._linear_terms_var) != len(other._linear_terms_var):
            return False

        # Establish a mapping between self's linear terms and other's
        linear_var_indices = {}
        for i in self._linear_terms_var:
            for j in other._linear_terms_var:
                if self._linear_terms_var[i] is other._linear_terms_var[j]:
                    linear_var_indices[i] = j

        # We have to have found all our own vars
        if len(linear_var_indices) != len(self._linear_terms_var):
            return False

        # Check that linear term coefficients are equal
        for i in self._linear_terms_var:
            if self._linear_terms_coef[i] != other._linear_terms_coef[linear_var_indices[i]]:
                return False

        if self._nonlinear_expr is not None:
            if other._nonlinear_expr is None:
                return False

            if type(self._nonlinear_expr) is list:
                if type(other._nonlinear_expr) is not list:
                    return False

                # Check that nonlinear expressions are the same length
                if len(self._nonlinear_expr) != len(other._nonlinear_expr):
                    return False

                # Check that nonlinear expression have the same terms
                for i in xrange(len(self._nonlinear_expr)):
                    if self._nonlinear_expr[i][0] != other._nonlinear_expr[i][0]:
                        return False
                    if not self._nonlinear_expr[i][1] is other._nonlinear_expr[i][1]:
                        return False
            else:
                if not self._nonlinear_expr is other._nonlinear_expr:
                    return False

        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def is_fixed(self):
        if len(self._linear_terms_var) == 0 and len(self._nonlinear_vars) == 0:
            return True
        return False

    def is_linear(self):
        if self._nonlinear_expr is None:
            return True

    def is_nonlinear(self):
        if self._nonlinear_expr is None:
            return False
        return True

    # GAH: This is not being used anywhere. What is it for?
    def needs_sum(self):
        num_components = 0
        if self._constant != 0:
            num_components += 1

        num_components += len(self._linear_terms_var)

        if not self._nonlinear_expr is None:
            num_components += 1

        if num_components > 1:
            return True
        return False

    # GAH: This is not being used anywhere. What is it for?
    def create_expr(self):
        if self.needs_sum() is True:
            ret_expr = expr._SumExpression()
            if self.is_fixed():
                ret_expr._const = self._constant

            for (var_name, var) in self._linear_terms_var.iteritems():
                ret_expr.add(self._linear_terms_coef[var_name], var)

            if self.is_nonlinear():
                ret_expr.add(1.0, self._nonlinear_expr)

            return ret_expr
        else:
            # we know there is only one component present in ampl_repn
            if self.is_fixed():
                return as_numeric(self._constant)
            elif self.is_nonlinear():
                return self._nonlinear_expr
            else:
                assert( len(self._linear_terms_var) == 1 )
                return self._linear_terms_var[self._linear_terms_var.keys()[0]]


def generate_ampl_repn(exp, symbol_map, labeler):
    ampl_repn = ampl_representation()
    #
    # Variable
    #
    if isinstance(exp, _VarData):
        if exp.fixed:
            ampl_repn._constant = exp.value
            return ampl_repn
        use_name = symbol_map.getSymbol( exp, labeler )
        ampl_repn._name_prefix = use_name[:len(use_name)-len(exp.name)]
        ampl_repn._linear_terms_coef[use_name] = 1.0
        ampl_repn._linear_terms_var[use_name] = exp
        return ampl_repn

    #
    # Expression
    #
    if isinstance(exp,expr.Expression):
        exp_type = type(exp)

        #
        # Sum
        #
        if exp_type is expr._SumExpression:
            ampl_repn._constant = exp._const #float(exp._const)
            ampl_repn._nonlinear_expr = None
            for i in xrange(len(exp._args)):
                exp_coef = exp._coef[i]
                child_repn = generate_ampl_repn(exp._args[i], symbol_map, labeler)
                if ampl_repn._name_prefix == "":
                    ampl_repn._name_prefix = child_repn._name_prefix
                # adjust the constant
                ampl_repn._constant += exp_coef*child_repn._constant

                # adjust the linear terms
                for (var_name,var) in child_repn._linear_terms_var.iteritems():
                    if var_name in ampl_repn._linear_terms_var:
                        ampl_repn._linear_terms_coef[var_name] += exp_coef*child_repn._linear_terms_coef[var_name]
                    else:
                        ampl_repn._linear_terms_var[var_name] = var
                        ampl_repn._linear_terms_coef[var_name] = exp_coef*child_repn._linear_terms_coef[var_name]

                # adjust the nonlinear terms
                if not child_repn._nonlinear_expr is None:
                    if ampl_repn._nonlinear_expr is None:
                        ampl_repn._nonlinear_expr = [(exp_coef, child_repn._nonlinear_expr)]
                    else:
                        ampl_repn._nonlinear_expr.append((exp_coef, child_repn._nonlinear_expr))

                # adjust the nonlinear vars
                ampl_repn._nonlinear_vars.update(child_repn._nonlinear_vars)
            return ampl_repn

        #
        # Product
        #
        elif exp_type is expr._ProductExpression:
            #
            # Iterate through the denominator.  If they aren't all constants, then
            # simply return this expresion.
            #
            denom=1.0
            for e in exp._denominator:
                if e.is_fixed():
                    denom *= value(e)
                else:
                    ampl_repn._nonlinear_expr = exp
                    break
                if denom == 0.0:
                    print "Divide-by-zero error - offending sub-expression:"
                    e.pprint()
                    raise ZeroDivisionError

            if not ampl_repn._nonlinear_expr is None:
                # we have a nonlinear expression ... build up all the vars
                for e in exp._denominator:
                    arg_repn = generate_ampl_repn(e, symbol_map, labeler)
                    ampl_repn._nonlinear_vars.update(arg_repn._linear_terms_var)
                    ampl_repn._nonlinear_vars.update(arg_repn._nonlinear_vars)

                for e in exp._numerator:
                    arg_repn = generate_ampl_repn(e, symbol_map, labeler)
                    ampl_repn._nonlinear_vars.update(arg_repn._linear_terms_var)
                    ampl_repn._nonlinear_vars.update(arg_repn._nonlinear_vars)
                if ampl_repn._name_prefix == "":
                    ampl_repn._name_prefix = arg_repn._name_prefix
                return ampl_repn

            #
            # OK, the denominator is a constant.
            #
            # build up the ampl_repns for the numerator
            
            ###### GAH: Deactivating this preprocessing step, which 
            ######      also eliminates the need for expression cloning
            # preprocess out the denominator
            #exp._denominator = []
            #exp.coef *= 1.0/denom
            # Make sure to reset the denom to 1.0, this is important
            # when the outer expression is linear so that
            # we don't propigate this effect into the ampl
            # repn more than once.
            #denom = 1.0
            ######

            n_linear_args = 0
            n_nonlinear_args = 0
            arg_repns = list()
            for i in xrange(len(exp._numerator)):
                e = exp._numerator[i]
                e_repn = generate_ampl_repn(e, symbol_map, labeler)
                arg_repns.append(e_repn)
                # check if the expression is not nonlinear else it is nonlinear
                if e_repn._nonlinear_expr is not None:
                    n_nonlinear_args += 1
                # Check whether the expression is constant or else it is linear
                elif len(e_repn._linear_terms_var) > 0:
                    n_linear_args += 1
                # At this point we do not have a nonlinear expression and there are
                # no linear terms. If the expression constant is zero, then we
                # have a zero term in the product expression, so the entire product 
                # expression becomes trivial. 
                elif e_repn._constant == 0.0:
                    ampl_repn = e_repn
                    return ampl_repn

            is_nonlinear = False
            if n_linear_args > 1 or n_nonlinear_args > 0:
                is_nonlinear = True


            if is_nonlinear is True:
                # do like AMPL and simply return the expression
                # without extracting the potentially linear part
                ampl_repn = ampl_representation()
                ampl_repn._nonlinear_expr = exp
                for repn in arg_repns:
                    ampl_repn._nonlinear_vars.update(repn._linear_terms_var)
                    ampl_repn._nonlinear_vars.update(repn._nonlinear_vars)
                if ampl_repn._name_prefix == "":
                    ampl_repn._name_prefix = repn._name_prefix
                return ampl_repn

            else: # is linear or constant
                ampl_repn = current_repn = arg_repns[0]
                for i in xrange(1,len(arg_repns)):
                    e_repn = arg_repns[i]
                    ampl_repn = ampl_representation()

                    # const_c * const_e
                    ampl_repn._constant = current_repn._constant * e_repn._constant

                    # const_e * L_c
                    if e_repn._constant != 0.0:
                        for (var_name, var) in current_repn._linear_terms_var.iteritems():
                            ampl_repn._linear_terms_coef[var_name] = current_repn._linear_terms_coef[var_name] * e_repn._constant
                            ampl_repn._linear_terms_var[var_name] = var

                    # const_c * L_e
                    if current_repn._constant != 0.0:
                        for (e_var_name,e_var) in e_repn._linear_terms_var.iteritems():
                            if e_var_name in ampl_repn._linear_terms_var:
                                ampl_repn._linear_terms_coef[e_var_name] += current_repn._constant * e_repn._linear_terms_coef[e_var_name]
                            else:
                                ampl_repn._linear_terms_coef[e_var_name] = current_repn._constant * e_repn._linear_terms_coef[e_var_name]
                                ampl_repn._linear_terms_var[e_var_name] = e_var
                    if ampl_repn._name_prefix == "":
                        ampl_repn._name_prefix = e_repn._name_prefix
                    current_repn = ampl_repn

            # now deal with the product expression's coefficient that needs
            # to be applied to all parts of the ampl_repn
            ampl_repn._constant *= exp.coef/denom
            for var_name in ampl_repn._linear_terms_coef:
                ampl_repn._linear_terms_coef[var_name] *= exp.coef/denom

            return ampl_repn

        #
        # Power Expressions
        #
        elif exp_type is expr._PowExpression:
            assert(len(exp._args) == 2)
            base_repn = generate_ampl_repn(exp._args[0], symbol_map, labeler)
            exponent_repn = generate_ampl_repn(exp._args[1], symbol_map, labeler)

            if base_repn.is_fixed() and exponent_repn.is_fixed():
                ampl_repn._constant = base_repn._constant**exponent_repn._constant
            elif exponent_repn.is_fixed() and exponent_repn._constant == 1.0:
                ampl_repn = base_repn
            elif exponent_repn.is_fixed() and exponent_repn._constant == 0.0:
                ampl_repn._constant = 1.0
            else:
                # instead, let's just return the expression we are given and only
                # use the ampl_repn for the vars
                ampl_repn._nonlinear_expr = exp
                ampl_repn._nonlinear_vars = base_repn._nonlinear_vars
                ampl_repn._nonlinear_vars.update(exponent_repn._nonlinear_vars)
                ampl_repn._nonlinear_vars.update(base_repn._linear_terms_var)
                ampl_repn._nonlinear_vars.update(exponent_repn._linear_terms_var)
            if ampl_repn._name_prefix == "":
                if len(base_repn._name_prefix) > len(exponent_repn._name_prefix):
                    ampl_repn._name_prefix = base_repn._name_prefix
                else:
                    ampl_repn._name_prefix = exponent_repn._name_prefix
            return ampl_repn

        #
        # Intrinsic Functions
        #
        elif exp_type is expr._IntrinsicFunctionExpression:
            assert(len(exp._args) == 1)
            
            # the argument is fixed, we can simply evaluate this expression
            if exp._args[0].is_fixed():
                ampl_repn._constant = value(exp)
                return ampl_repn

            # this is inefficient since it is using much more than what we need
            child_repn = generate_ampl_repn(exp._args[0], symbol_map, labeler)

            ampl_repn._nonlinear_expr = exp
            ampl_repn._nonlinear_vars = child_repn._nonlinear_vars
            ampl_repn._nonlinear_vars.update(child_repn._linear_terms_var)
            if ampl_repn._name_prefix == "":
                ampl_repn._name_prefix = child_repn._name_prefix
            return ampl_repn
        #
        # ERROR
        #
        else:
            raise ValueError, "Unsupported expression type: "+str(exp)
    #
    # Constant
    #
    elif exp.is_fixed():
        ### GAH: Why were we even checking this
        #if not exp.value.__class__ in native_numeric_types:
        #    ampl_repn = generate_ampl_repn(exp.value, symbol_map, labeler) 
        #    return ampl_repn
        ampl_repn._constant = value(exp)
        return ampl_repn

    #
    # ERROR
    #
    else:
        raise ValueError, "Unexpected expression type: "+str(exp)



# Switch from Python to C generate_ampl_repn function when possible
try:
    py_generate_ampl_repn = generate_ampl_repn
    from cAmpl import generate_ampl_repn
except ImportError:
    del py_generate_ampl_repn

generate_ef_ampl_repn = generate_ampl_repn # legacy nam

# Alternative: import C implementation under another name for testing
#from cAmpl import generate_ampl_repn as cgar
#__all__.append('cgar')



"""
This example shows how to represent a piecewise function using 
Pyomo's built SOSConstraint component. The function is defined as:

       / 3x-2 , 1 <= x <= 2
f(x) = | 
       \ 5x-6 , 2 <= x <= 3
"""

from coopr.pyomo import *

INDEX_SET = [1,2] # There will be two copies of this function
DOMAIN_PTS = {1:[1,2,3], 2:[1,2,3]}
F = {1:[1,4,9],2:[1,4,9]}
# Note we can also implement this like below
#F = lambda x: x**2
# Update the return value for constraint2_rule if
# F is defined using the function above

model = ConcreteModel()

# Indexing set required for the SOSConstraint declaration
def SOS_indices_init(model,t):
    return [(t,i) for i in xrange(len(DOMAIN_PTS[t]))]
model.SOS_indices = Set(INDEX_SET,dimen=2, ordered=True, initialize=SOS_indices_init)

def sos_var_indices_init(model):
    return [(t,i) for t in INDEX_SET for i in xrange(len(DOMAIN_PTS[t]))]
model.sos_var_indices = Set(ordered=True, dimen=2,initialize=sos_var_indices_init)

model.x = Var(INDEX_SET) # domain variable
model.Fx = Var(INDEX_SET) # range variable
model.y = Var(model.sos_var_indices,within=NonNegativeReals) # SOS2 variable

model.obj = Objective(expr=summation(model.Fx), sense=maximize)

def constraint1_rule(model,t):
    return model.x[t] == sum(model.y[t,i]*DOMAIN_PTS[t][i] for i in xrange(len(DOMAIN_PTS[t])) )
def constraint2_rule(model,t):
    # Uncomment below for F defined as dictionary
    return model.Fx[t] == sum(model.y[t,i]*F[t][i] for i in xrange(len(DOMAIN_PTS[t])) )
    # Uncomment below for F defined as lambda function
    #return model.Fx[t] == sum(model.y[t,i]*F(DOMAIN_PTS[t][i]) for i in xrange(len(DOMAIN_PTS[t])) )
def constraint3_rule(model,t):
    return sum(model.y[t,j] for j in xrange(len(DOMAIN_PTS[t]))) == 1

model.constraint1 = Constraint(INDEX_SET,rule=constraint1_rule)
model.constraint2 = Constraint(INDEX_SET,rule=constraint2_rule)
model.constraint3 = Constraint(INDEX_SET,rule=constraint3_rule)
model.SOS_set_constraint = SOSConstraint(INDEX_SET, var=model.y, set=model.SOS_indices, sos=2)

#Fix the answer for testing purposes
model.set_answer_constraint1 = Constraint(expr= model.x[1] == 2.5)
model.set_answer_constraint2 = Constraint(expr= model.x[2] == 2.0)



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

import log_config
from numvalue import *
from expr import *
from intrinsic_functions import *
from label import *
from plugin import *
from PyomoModelData import *
#
# Components
#
from component import *
from sets import *
from param import *
from var import *
from constraint import *
from objective import *
from connector import *
from sos import *
from piecewise import *
#
from set_types import *
from misc import *
from block import *
from PyomoModel import *
#
import pyomo
#
from util import *
from rangeset import *

from instance2dat import *

#
# This is a hack to strip out modules, which shouldn't have been included in these imports
#
import types
_locals = locals()
__all__ = [__name for __name in _locals.keys() if (not __name.startswith('_') and not isinstance(_locals[__name],types.ModuleType)) or __name == '_' ]
__all__.append('pyomo')



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

import os.path
import re
from pyutilib.misc import Options
from pyutilib.component.core import alias, Plugin, implements
from coopr.pyomo.base.plugin import IDataManager
from process_data import _process_include


class AmplDataCommands(Plugin):

    alias("dat", "Import data with AMPL data commands.")

    implements(IDataManager)

    def __init__(self):
        self._info = []

    def initialize(self, filename, **kwds):
        self.filename = filename
        self.options = Options(**kwds)

    def open(self):
        if self.filename is None:
            raise IOError, "No filename specified"
        if not os.path.exists(self.filename):
            raise IOError, "Cannot find file '%s'" % self.filename

    def close(self):
        pass

    def read(self):
        """
        This function does nothing, since executing AMPL data commands
        both reads and processes the data all at once.
        """
        return True

    def process(self, model, data, default):
        """
        Read AMPL data commands and process the data.
        """
        _process_include(['include', self.filename], model, data, default)

    def clear(self):
        self._info = []



from coopr.pyomo import *

PIECEWISE_PTS = [1,2,3]
F = lambda x: x**2

def define_model():
    model = ConcreteModel()

    def SOS_indices_init(model):
        return [i for i in xrange(1,len(PIECEWISE_PTS))]
    def ub_indices_init(model):
        return [i for i in xrange(1,len(PIECEWISE_PTS))]
    
    model.SOS_indices = Set(dimen=1, ordered=True, initialize=SOS_indices_init)
    model.ub_indices = Set(ordered=True, dimen=1,initialize=ub_indices_init)

    model.x = Var()
    model.Fx = Var()
    #Add SOS1 variable to model
    model.y = Var(model.ub_indices,within=NonNegativeReals)

    def constraint1_rule(model,i):
        return model.Fx - (F(PIECEWISE_PTS[i-1]) + ((F(PIECEWISE_PTS[i])-F(PIECEWISE_PTS[i-1]))/(PIECEWISE_PTS[i]-PIECEWISE_PTS[i-1]))*(model.x-PIECEWISE_PTS[i-1])) <= 25.0*(1-model.y[i])
    def constraint2_rule(model):
        return sum(model.y[j] for j in xrange(1,len(PIECEWISE_PTS))) == 1

    model.obj = Objective(expr=summation(model.Fx), sense=maximize)
    model.constraint1 = Constraint(model.ub_indices,rule=constraint1_rule)
    model.constraint2 = Constraint(rule=constraint2_rule)
    model.SOS_set_constraint = SOSConstraint(var=model.y, set=model.SOS_indices, sos=1)

    #Fix the answer for testing purposes
    model.set_answer_constraint1 = Constraint(expr= model.x == 2.5)

    return model

model = define_model()



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

__all__ = ['SparseIndexedComponent']

import weakref
import pyutilib.misc

import component
from sets import Set
from indexed_component import IndexedComponent

def process_setarg(arg):
    """
    Process argument and return an associated set object.
    """
    if isinstance(arg,Set):
        # Argument is a Set instance
        return arg
    elif isinstance(arg,IndexedComponent) or \
            isinstance(arg,SparseIndexedComponent):
        raise TypeError, "Cannot index a component with a non-set component"
    else:
        try:
            # Argument has set_options attribute, which is used to initialize the set
            options = getattr(arg,'set_options')
            options['initialize'] = arg
            return Set(**options)
        except:
            pass
    # Argument is assumed to be an initialization function
    return Set(initialize=arg)


UnindexedComponent_set = set([None])

class SparseIndexedComponent(component.Component):
    """
    This is the base class for all indexed modeling components.

    Constructor arguments:
        ctype       The class type for the derived subclass
        doc         A text string describing this component

    Private class attributes:
        _data       A dictionary from the index set to component data objects
        _index      The set of valid indices
    """

    def __init__(self, *args, **kwds):
        component.Component.__init__(self, **kwds)
        #
        self._data = {}
        #
        if len(args) == 0:
            #
            # If no indexing sets are provided, generate a dummy index
            #
            self._implicit_subsets = None
            self._index = UnindexedComponent_set
            self._data = {None: weakref.ref(self)}
        elif len(args) == 1:
            self._implicit_subsets = None
            self._index = process_setarg(args[0])
        else:
            # NB: Pyomo requires that all modelling components are
            # assigned to the model.  The trick is that we allow things
            # like "Param([1,2,3], range(100), initialize=0)".  This
            # needs to create *3* sets: two SetOf components and then
            # the SetProduct.  That means that teh component needs to
            # hold on to the implicit SetOf objects until the component
            # is assigned to a model (where the implicit subsets can be
            # "transferred" to the model).
            tmp = [process_setarg(x) for x in args]
            self._implicit_subsets = tmp
            self._index = tmp[0].cross(*tmp[1:])
        #
        # Setup indices for suffix data
        #
        self._suffix_declaration = {}
        self._suffix_value = {}

    def clear(self):
        """Clear the data in this component"""
        if UnindexedComponent_set != self._index:
            self._data = {}
        else:
            raise NotImplementedError(
                "Derived singleton component %s failed to define clear().\n"
                "\tPlease report this to the Pyomo developers"
                % (self.__class__.__name__,))

    def __len__(self):
        return len(self._index)

    def __contains__(self, ndx):
        return ndx in self._index

    #
    # NB: The standard access / iteration methods treat this component
    # as a dense set.  For some sparse components, this may result in
    # "filling in" the sparse matrix (through the derived class'
    # _default(idx) method).
    #
    
    def __iter__(self):
        return self._index.__iter__()

    def keys(self):
        # Leverage the iterator (generator) to generate the dense 
        return [ x for x in self ]

    def values(self):
        # Leverage the iterator (generator) to generate the dense 
        return [ self[x] for x in self ]

    def items(self):
        # Leverage the iterator (generator) to generate the dense 
        return [ (x, self[x]) for x in self ]

    def iterkeys(self):
        return self._index.__iter__()

    def itervalues(self):
        for key in self:
            yield self[key]
    
    def iteritems(self):
        for key in self:
            yield (key, self[key])

    #
    # NB: These are "sparse equivalent" access / iteration methods that
    # only loop over the defined data.  This preserves the object's
    # sparsity, but explicitly excludes non-defined values.
    #

    def sparse_keys(self):
        return self._data.keys()

    def sparse_values(self):
        return [ self[x] for x in self._data.keys() ]

    def sparse_items(self):
        return [ (x, self[x]) for x in self._data.keys() ]

    def sparse_iterkeys(self):
        return self._data.iterkeys()

    def sparse_itervalues(self):
        for key in self._data.iterkeys():
            yield self[key]
    
    def sparse_iteritems(self):
        for key in self._data.iterkeys():
            yield (key, self[key])


    def __getitem__(self, ndx):
        """
        This method returns the data corresponding to the given index.
        """
        if ndx in self._data:
            if ndx is None:
                return self
            else:
                return self._data[ndx]
        elif ndx in self._index:
            return self._default(ndx)
        else:
            ndx = self.normalize_index(ndx)
            if ndx in self._data:
                if ndx is None:
                    return self
                else:
                    return self._data[ndx]
            if ndx in self._index:
                return self._default(ndx)

        if not self.is_indexed():
            msg = "Error accessing indexed component: " \
                  "Cannot treat the scalar component '%s' as an array" \
                  % ( self.name, )
        else:
            msg = "Error accessing indexed component: " \
                  "Index '%s' is not valid for array component '%s'" \
                  % ( ndx, self.name, )
        raise KeyError(msg)

    def _default(self, index):
        raise NotImplementedError(
            "Derived component %s failed to define _default().\n"
            "\tPlease report this to the Pyomo developers"
            % (self.__class__.__name__,))

    def normalize_index(self, index):
        ndx = pyutilib.misc.flatten(index)
        if type(ndx) is list:
            if len(ndx) == 1:
                ndx = ndx[0]
            else:
                ndx = tuple(ndx)
        return ndx
            

    def index(self):
        """Return the index set"""
        return self._index

    def is_indexed(self):
        """Return true if this component is indexed"""
        return UnindexedComponent_set != self._index

    def dim(self):
        """Return the dimension of the index"""
        if UnindexedComponent_set != self._index:
            return self._index.dimen
        else:
            return 0

    def set_value(self, value):
        if UnindexedComponent_set != self._index:
            raise ValueError(
                "Cannot set the value for the indexed component '%s' "
                "without specifying an index value.\n"
                "\tFor example, model.%s[i] = value"
                % (self.name, self.name))
        else:
            raise NotImplementedError(
                "Derived component %s failed to define set_value() "
                "for singleton instances.\n"
                "\tPlease report this to the Pyomo developers"
                % (self.__class__.__name__,))
            

    def declare_suffix(self, suffix, default=None, rule=None):
        """
        Declare a suffix attribute.
        """
        if suffix[0] == "_":
            raise AttributeError(
                "Cannot define a suffix that begins with  '_'")
        if suffix in self._suffix_declaration:
            raise AttributeError("Suffix %s is already defined" % suffix)
        self._suffix_declaration[suffix] = (default,rule)
        #
        # If this variable has been constructed, then
        # generate this attribute for all variables
        #
        #if len(self._data) > 0:
            #self._suffix_value[suffix] = {}

    def suffix_defined(self, suffix):
        """
        Determine if a suffix has aready been declared.
        """
        return suffix in self._suffix_declaration

    def get_suffix_value(self, suffix, index=None):
        try:
            return self._suffix_value[suffix, index]
        except KeyError:
            default = self._suffix_declaration[suffix]
            if not default[0] is None:
                return default[0]
            if not default[1] is None:
                val = default[1](self)
                self._suffix_value[suffix, index] = val
                return val
            return None

    def set_suffix_value(self, suffix, value, index=None, define=False):
        if define and suffix not in self._suffix_declaration:
            self.declare_suffix(suffix)
        self._suffix_value[suffix, index] = value




#
# Unit Tests for model preprocessing
#

import os
import sys
from os.path import abspath, dirname
sys.path.insert(0, dirname(dirname(abspath(__file__)))+"/../..")
currdir = dirname(abspath(__file__))+os.sep

import pyutilib.th as unittest
from coopr.pyomo import *
from nose.tools import nottest


class TestPreprocess(unittest.TestCase):

    def Xtest_label1(self):
        model = AbstractModel()
        model.A = Set(initialize=[1,2,3])
        model.B = Param(model.A,initialize={1:100,2:200,3:300})
        model.x = Var(model.A)
        model.y = Var(model.A)
        instance=model.create()
        instance.preprocess()
        self.assertEqual(instance.num_used_variables(),0)

    def Xtest_label2(self):
        model = AbstractModel()
        model.A = Set(initialize=[1,2,3])
        model.B = Param(model.A,initialize={1:100,2:200,3:300})
        model.x = Var(model.A)
        model.y = Var(model.A)
        model.obj = Objective(rule=lambda inst: inst.x[1])
        instance=model.create()
        instance.preprocess()
        self.assertEqual(instance.num_used_variables(),1)
        self.assertEqual(instance.x[1].label,"x(1)")
        self.assertEqual(instance.x[2].label,"x(2)")
        self.assertEqual(instance.y[1].label,"y(1)")

if __name__ == "__main__":
    unittest.main()



#
# Imports
#
import sys
sys.path.append("../../..")
from coopr.pyomo import *
#pyomo.set_debugging()

##
## Using a Model
##
#
# Pyomo makes a fundamental distinction between an abstract model and a
# problem instance.  The Pyomo AbstractModel() class is used to manage the 
# declaration of model components (e.g. sets and variables), and to 
# generate a problem instance.
#
model = AbstractModel()

##
## Declaring Sets
##
#
# An unordered set of arbitrary objects
#
model.A = Set()
#
# An unordered set of numeric values
#
model.B = Set()
#
# A simple cross-product
#
model.C = model.A * model.B
#
# A simple cross-product loaded with a tabular data format
#
model.D = Set(within=model.A * model.B)
#
# A multiple cross-product
#
model.E = Set(within=model.A * model.B * model.A)

#
# An indexed set
# 
model.F = Set(model.A)
#
# An indexed set
# 
model.G = Set(model.A,model.B)
#
# A simple set
#
model.H = Set()
#
# A simple set
#
model.I = Set()
#
# A two-dimensional set
#
model.J = Set(dimen=2)

##
## Declaring Params
##
#
#
# A simple parameter
#
model.Z = Param()
#
# A single-dimension parameter
#
model.Y = Param(model.A)
#
# An example of initializing two single-dimension parameters together
#
model.X = Param(model.A)
model.W = Param(model.A)
#
# Initializing a parameter with two indices
#
model.U = Param(model.I,model.A)
model.T = Param(model.A,model.I)
#
# Initializing a parameter with missing data
#
model.S = Param(model.A)
#
# An example of initializing two single-dimension parameters together with
# an index set
#
model.R = Param(model.H, within=Reals)
model.Q = Param(model.H, within=Reals)
#
# An example of initializing parameters with a two-dimensional index set
#
model.P = Param(model.J, within=Reals)
model.PP = Param(model.J, within=Reals)
model.O = Param(model.J, within=Reals)
   
##
## Process an input file and confirm that we get appropriate 
## set instances.
##
#model.pprint()

data = ModelData(model=model)
data.add("tab/A.tab", format='set', set='A')
data.add("tab/B.tab", format='set', set='B')
data.add("tab/C.tab", format='set', set="C")
data.add("tab/D.tab", format="set_array", set='D')
data.add("tab/E.tab", format='set', set="E")
data.add("tab/I.tab", format='set', set='I')
data.add("tab/Z.tab", format='param', param="Z")
data.add("tab/Y.tab", index=['A'], param=['Y'])
data.add("tab/XW.tab", index=['A'], param=['X','W'])
data.add("tab/T.tab", param="T", format="transposed_array")
data.add("tab/U.tab", param="U", format="array")
data.add("tab/S.tab", index=['A'], param=['S'])
data.add("tab/RQ.tab", index_name="H", index=['H'], param=("R","Q"))
data.add("tab/PO.tab", index_name="J", index=('A','B'), param=("P","O"))
data.add("tab/PP.tab", index=('A','B'), param="PP")
try:
    data.read()
except coopr.ApplicationError:
    sys.exit(0)

instance = model.create(data)
instance.pprint()




import argparse
import coopr.misc.coopr_parser
import os.path

class EnableDisableAction(argparse.Action):
    def add_package(self, namespace, package):
        if namespace.checkers.get(package, None) is None:
            namespace.checkers[package] = []
        for c in coopr.pyomo.check.ModelCheckRunner._checkers(all=True):
            if c._checkerPackage() == package:
                namespace.checkers[package].append(c._checkerName())

    def remove_package(self, namespace, package):
        if package in namespace.checkers:
            del namespace.checkers[package]

    def add_checker(self, namespace, checker):
        for c in coopr.pyomo.check.ModelCheckRunner._checkers(all=True):
            if c._checkerName() == checker:
                if namespace.checkers.get(c._checkerPackage(), None) is None:
                    namespace.checkers[c._checkerPackage()] = []
                if c._checkerName() not in namespace.checkers[c._checkerPackage()]:
                    namespace.checkers[c._checkerPackage()].append(c._checkerName())

    def remove_checker(self, namespace, checker):
        for c in coopr.pyomo.check.ModelCheckRunner._checkers(all=True):
            if c._checkerName() == checker:
                if namespace.checkers.get(c._checkerPackage(), None) is not None:
                    for i in range(namespace.checkers[c._checkerPackage()].count(c._checkerName())):
                        namespace.checkers[c._checkerPackage()].remove(c._checkerName())

    def add_default_checkers(self, namespace):
        self.add_package(namespace, 'pyomo')
        self.add_package(namespace, 'py3k')

    def __call__(self, parser, namespace, values, option_string=None):
        if 'checkers' not in dir(namespace):
            setattr(namespace, 'checkers', {})
            self.add_default_checkers(namespace)
        
        if option_string == '-c':
            self.add_checker(namespace, values)
        elif option_string == '-C':
            self.add_package(namespace, values)
        elif option_string == '-x':
            self.remove_checker(namespace, values)
        elif option_string == '-X':
            self.remove_package(namespace, values)

def setup_parser(parser):
    parser.add_argument("script", metavar="SCRIPT", default=None,
                        help="a Pyomo script that is checked")
    parser.add_argument("-v", "--verbose", action="store_true", dest="verbose",
                        default=False, help="enable additional output messages")
    parser.add_argument("-c", "--enable-checker", action=EnableDisableAction, 
                        help="activate a specific checker")
    parser.add_argument("-C", "--enable-package", action=EnableDisableAction,
                        help="activate an entire checker package")
    parser.add_argument("-x", "--disable-checker", action=EnableDisableAction,
                        help="disable a specific checker")
    parser.add_argument("-X", "--disable-package", action=EnableDisableAction,
                        help="disable an entire checker package")
    parser.add_argument("-s", "--show-checkers", action="store_true",
                        dest="show_checkers", help="show enabled checkers")


def main_exec(options):
    import coopr.pyomo.check as check

    if options.script is None:
        raise IOError, "Must specify a model script!"
    if not os.path.exists(options.script):
        raise IOError, "Model script '%s' does not exist!" % options.script

    # force default checkers
    if getattr(options, 'checkers', None) is None:
        EnableDisableAction(None, None)(None, options, None, None)

    runner = check.ModelCheckRunner()
    runner.run(**vars(options))

#
# Add a subparser for the check command
#
setup_parser(
    coopr.misc.coopr_parser.add_subparser('check',
        func=main_exec, 
        help='Check a model for errors.',
        description='This coopr subcommand is used to check a model script for errors.',
        epilog='The default behavior of this command is to assume that the model script is a simple Pyomo model.  Eventually, this script will support options that allow other Coopr models to be checked.'
        ))



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________


# Transformation heirarchy
from transformation import *
from linear_transformation import *
from nonlinear_transformation import *
from abstract_transformation import *
from concrete_transformation import *
from isomorphic_transformation import *
from nonisomorphic_transformation import *

# Transformations
import relax_integrality
import eliminate_fixed_vars
from standard_form import *
from equality_transform import *
from nonnegative_transform import *
from dual_transformation import DualTransformation
import util



from pyutilib.component.core import *

class IPreCheckHook(Interface):
    def precheck(self, runner, script, info):
        pass

class IPostCheckHook(Interface):
    def postcheck(self, runner, script, info):
        pass



from coopr.pyomo.check.checker import IterativeTreeChecker

class PrintASTNodes(IterativeTreeChecker):
    def __init__(self):
        self.disable()

    def check(self, runner, script, info):
        if 'lineno' in dir(info):
            self.problem(str(info), lineno = info.lineno)
        else:
            self.problem(str(info))



#
# Author:  Gabe Hackebeil
# Purpose: For regression testing to ensure that the Pyomo
#          NL writer properly reclassifies nonlinear expressions
#          as linear or trivial when fixing variables or params
#          cause such a situation.
#
#          This test model relies on the asl_test executable. It
#          will not solve if sent to a real optimizer.
#

from coopr.pyomo import *

model = ConcreteModel()

model.x = Var()
model.y = Var()
model.z = Var()
model.q = Param(initialize=0.0)
model.p = Param(initialize=0.0,mutable=True)

model.obj = Objective( expr=model.x*model.y +\
                            model.z*model.y +\
                            model.q*model.y +\
                            model.y*model.y*model.q +\
                            model.p*model.y +\
                            model.y*model.y*model.p +\
                            model.y*model.y*model.z +\
                            model.z*(model.y**2))

model.con1 = Constraint(expr=model.x*model.y == 0)
model.con2 = Constraint(expr=model.z*model.y + model.y == 0)
model.con3 = Constraint(expr=model.q*(model.y**2) + model.y == 0)
model.con4 = Constraint(expr=model.q*model.y*model.x + model.y == 0)
model.con5 = Constraint(expr=model.p*(model.y**2) + model.y == 0)
model.con6 = Constraint(expr=model.p*model.y*model.x + model.y == 0)
model.con7 = Constraint(expr=model.z*(model.y**2) + model.y == 0)
model.con8 = Constraint(expr=model.z*model.y*model.x + model.y == 0)
# Pyomo differs from AMPL in these cases that involve immutable params (q).
# These never actually become constraints in Pyomo, and for good reason.
model.con9 = Constraint(expr=model.z*model.y == 0)
model.con10 = Constraint(
    expr= simple_constraint_rule(model.q*(model.y**2) == 0) )
model.con11 = Constraint(
    expr= simple_constraint_rule(model.q*model.y*model.x == 0) )
model.con12 = Constraint(expr=model.p*(model.y**2) == 0)
model.con13 = Constraint(expr=model.p*model.y*model.x == 0)
model.con14 = Constraint(expr=model.z*(model.y**2) == 0)
model.con15 = Constraint(expr=model.z*model.y*model.x == 0)
model.con16 = Constraint(
    expr= simple_constraint_rule(model.q*model.y == 0) )
model.con17 = Constraint(expr=model.p*model.y == 0)

###### Add some constraint which we deactivate just
###### to make sure this is working properly
model.con1D = Constraint(expr=model.x*model.y == 0)
model.con1D_indexeda = Constraint([1,2],expr=lambda model,i: model.x*model.y == 0)
model.con1D_indexedb = Constraint([1,2],expr=lambda model,i: model.x*model.y == 0)
model.con1D.deactivate()
model.con1D_indexeda.deactivate()
model.con1D_indexedb[1].deactivate()
model.con1D_indexedb[2].deactivate()
#####


model.x = 1.0
model.x.fixed = True
model.z = 0.0
model.z.fixed = True



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________

"""
Script to generate the installer for coopr.pyomo.
"""

import glob
import os

def _find_packages(path):
    """
    Generate a list of nested packages
    """
    pkg_list=[]
    if not os.path.exists(path):
        return []
    if not os.path.exists(path+os.sep+"__init__.py"):
        return []
    else:
        pkg_list.append(path)
    for root, dirs, files in os.walk(path, topdown=True):
        if root in pkg_list and "__init__.py" in files:
            for name in dirs:
                if os.path.exists(root+os.sep+name+os.sep+"__init__.py"):
                    pkg_list.append(root+os.sep+name)
    return [pkg for pkg in map(lambda x:x.replace(os.sep,"."), pkg_list)]

def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()

from setuptools import setup
packages = _find_packages('coopr')

setup(name='coopr.pyomo',
      version='3.3.1',
      maintainer='William E. Hart',
      maintainer_email='wehart@sandia.gov',
      url = 'https://software.sandia.gov/svn/public/coopr/coopr.pyomo',
      license = 'BSD',
      platforms = ["any"],
      description = "Coopr's Pyomo math programming language",
      long_description = read('README.txt'),
      classifiers = [
            'Development Status :: 4 - Beta',
            'Intended Audience :: End Users/Desktop',
            'Intended Audience :: Science/Research',
            'License :: OSI Approved :: BSD License',
            'Natural Language :: English',
            'Operating System :: Microsoft :: Windows',
            'Operating System :: Unix',
            'Programming Language :: Python',
            'Programming Language :: Unix Shell',
            'Topic :: Scientific/Engineering :: Mathematics',
            'Topic :: Software Development :: Libraries :: Python Modules',
            'Topic :: Scientific/Engineering'
        ],
      packages=packages,
      keywords=['optimization'],
      namespace_packages=['coopr'],
      entry_points = """
        [coopr.command]
        coopr.check = coopr.pyomo.check.driver
        [console_scripts]
        pyomo=coopr.pyomo.scripting.pyomo:main
        pyomo2nl=coopr.pyomo.scripting.convert:pyomo2nl_main
        pyomo2lp=coopr.pyomo.scripting.convert:pyomo2lp_main
        pyomo2osil=coopr.pyomo.scripting.convert:pyomo2osil_main
      """
      )



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

__all__ = [ 'Connector' ]

import logging
import weakref

import sys

from block import Block
from component import Component
from constraint import Constraint, ConstraintList
from expr import _ProductExpression
from indexed_component import IndexedComponent
from numvalue import NumericValue
from plugin import IPyomoScriptModifyInstance
from var import VarList

from pyutilib.component.core import alias, Plugin, implements

logger = logging.getLogger('coopr.pyomo')

class _ConnectorValue(NumericValue):
    """Holds the actual connector information"""

    __slots__ = ('connector','index','vars','aggregators')

    def __init__(self, name):
        """Constructor"""

        # IMPT: The following three lines are equivalent to calling the
        #       basic NumericValue constructor, i.e., as follows:
        #       NumericValue.__init__(self, name, domain, None, False)
        #       That particular constructor call takes a lot of time
        #       for big models, and is unnecessary because we're not
        #       validating any values.
        self.name = name
        self.domain = None
        self.value = None

        # NOTE: the "name" attribute (part of the base NumericValue class) is
        #       typically something like some_var[x,y,z] - an easy-to-read
        #       representation of the variable/index pair.

        # NOTE: both of the following are presently set by the parent. arguably, we
        #       should provide keywords to streamline initialization.
        self.connector = None # the "parent" variable.
        self.index = None # the index of this variable within the "parent"
        self.vars = {}
        self.aggregators = {}
    
    def __str__(self):
        # the name can be None, in which case simply return "".
        if self.name is None:
            return ""
        else:
            return self.name

    def __getstate__(self):
        result = NumericValue.__getstate__(self)
        for i in _ConnectorValue.__slots__:
            result[i] = g_varvalueetattr(self, i)
        if type(result['connector']) is weakref.ref:
            result['connector'] = result['connector']()
        return result

    def __setstate__(self, dict):
        for (slot_name, value) in state.iteritems():
            self.__dict__[slot_name] = value
        if 'connector' in self.__dict__.keys() and self.connector is not None:
            self.connector = weakref.ref(self.connector)

    def set_value(self, value):
        msg = "Cannot specify the value of a connector '%s'"
        raise ValueError, msg % self.name

    def is_fixed(self):
        if len(self.vars) == 0:
            return False
        for var in self.vars.itervalues():
            if not var.is_fixed():
                return False
        return True

    def is_constant(self):
        for var in self.vars.itervalues():
            if not var.is_constant():
                return False
        return True

    def polynomial_degree(self):
        if self.is_fixed():
            return 0
        return 1

    def is_binary(self):
        for var in self.vars.itervalues():
            if var.is_binary():
                return True
        return False

    def is_integer(self):
        for var in self.vars.itervalues():
            if var.is_integer():
                return True
        return False

    def is_continuous(self):
        for var in self.vars.itervalues():
            if var.is_continuous():
                return True
        return False

    def pprint(self, ostream=None, verbose=False):
        if ostream is None:
            ostream = sys.stdout
        print >>ostream, str(self),

    def add(self, var, name=None, aggregate=None):
        if name is None:
            name = var.name
        if name in self.vars:
            raise ValueError("Cannot insert duplicate variable name "
                             "'%s' into Connector '%s'" % ( name, self.name ))
        self.vars[name] = var
        if aggregate is not None:
            self.aggregators[var] = aggregate


class _ConnectorBase(IndexedComponent):
    """A collection of variables, which may be defined over a index"""

    """ Constructor
        Arguments:
           name         The name of this connector
           index        The index set that defines the distinct connectors.
                          By default, this is None, indicating that there
                          is a single connector.
    """

    # TODO: default and rule keywords are not used?  Need to talk to Bill ...?
    def __init__(self, *args, **kwd):
        kwd.setdefault('ctype', Connector)
        self._initialize = kwd.pop('initialize', None)
        IndexedComponent.__init__(self, *args, **kwd)
        self._conval = {}

    def as_numeric(self):
        if None in self._conval:
            return self._conval[None]
        return self

    def is_indexed(self):
        return self._ndim > 0

    def is_expression(self):
        return False

    def is_relational(self):
        return False

    def keys(self):
        return self._conval.keys()

    def __iter__(self):
        return self._conval.keys().__iter__()

    def iteritems(self):
        return self._conval.iteritems()

    def __contains__(self,ndx):
        return ndx in self._conval

    def dim(self):
        return self._ndim

    def __len__(self):
        return len(self._conval)

    def __getitem__(self,ndx):
        """This method returns a _ConnectorValue object.
        """
        try:
            return self._conval[ndx]
        except KeyError: # thrown if the supplied index is hashable, but not defined.
            msg = "Unknown index '%s' in connector %s;" % (str(ndx), self.name)
            if (isinstance(ndx, (tuple, list)) and len(ndx) != self.dim()):
                msg += "    Expecting %i-dimensional indices" % self.dim()
            else:
                msg += "    Make sure the correct index sets were used.\n"
                msg += "    Is the ordering of the indices correct?"
            raise KeyError, msg
        except TypeError, msg: # thrown if the supplied index is not hashable
            msg2 = "Unable to index connector %s using supplied index with " % self.name
            msg2 += str(msg)
            raise TypeError, msg2

    def _add_indexed_member(self, ndx):
        new_conval = _ConnectorValue(create_name(self.name,ndx))
        new_conval.component = weakref.ref(self)
        new_conval.index = ndx
        
        self._conval[ndx] = new_conval

    def construct(self, data=None):
        if __debug__:   #pragma:nocover
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("Constructing Connector, name=%s, from data=%s", self.name, `data`)
        if self._constructed:
            return
        self._constructed=True
        #
        # Construct _ConnectorValue objects for all index values
        #
        init = self._initialize is not None
        if self._ndim > 0:
            for ndx in self._index:
                self._add_indexed_member(ndx)
                tmp = self[ndx]
                if init:
                    for key, val in self._initialize.iteritems():
                        tmp.add(val,key)
        else:
            # if the dimension is a singleton (i.e., we're dealing
            # with a _ConElement), and the _data is already initialized.
            if init:
                for key, val in self._initialize.iteritems():
                    self.add(val,key)
            

    def pprint(self, ostream=None, verbose=False):
        if ostream is None:
            ostream = sys.stdout
        print >>ostream, "  ",self.name,":",
        print >>ostream, "\tSize="+str(len(self)),
        if self._index_set is not None:
            print >>ostream, "\tIndicies: ",
            for idx in self._index_set:
                print >>ostream, str(idx.name)+", ",
            print ""
        if None in self._conval:
            print >>ostream, "\tName : Variable"
            for item in self._conval[None].iteritems():
                print >>ostream, "\t %s : %s" % item
        else:
            print >>ostream, "\tKey : Name : Variable"
            tmp=self._conval.keys()
            tmp.sort()
            for key in tmp:
                for name, var in self._conval[key].iteritems():
                    print >>ostream, "\t %s : %s : %s" % ( key, name, var )


    def display(self, prefix="", ostream=None):
        if ostream is None:
            ostream = sys.stdout
        print >>ostream, prefix+"Connector "+self.name,":",
        print >>ostream, "  Size="+str(len(self)),
        if None in self._conval:
            print >>ostream, prefix+"  : {"+\
                ', '.join(sorted(self._conval[key].keys()))+"}"
        else:
            for key in sorted(self._conval.keys()):
                print >>ostream, prefix+"  "+str(key)+" : {"+\
                  ', '.join(sorted(self._conval[key].keys()))+"}"


class _ConnectorElement(_ConnectorBase, _ConnectorValue):

    def __init__(self, *args, **kwd):

        _ConnectorValue.__init__(self, kwd.get('name', None) )
        _ConnectorBase.__init__(self, *args, **kwd)
        self._conval[None] = self
        self._conval[None].component = weakref.ref(self)
        self._conval[None].index = None

    def __getstate__(self):
        result = _VarData.__getstate__(self)
        for key,value in self.__dict__.iteritems():
            result[key]=value
        if type(result['_conval'][None].component) is weakref.ref:
            result['_conval'][None].component = None
        return result

    def __setstate__(self, dict):
        for key in dict:
            setattr(self, key, dict[key])
        self._conval[None].component = weakref.ref(self)

    def is_constant(self):
        return _ConnectorValue.is_constant(self)


# a _ConnectorArray is the implementation representing an indexed connector.

class _ConnectorArray(_ConnectorBase):
    
    def __init__(self, *args, **kwds):

        _ConnectorBase.__init__(self, *args, **kwds)
        self._dummy_val = _ConnectorValue(kwds.get('name', None))

    def __float__(self):
        raise TypeError, "Cannot access the value of array connector "+self.name

    def __int__(self):
        raise TypeError, "Cannot access the value of array connector "+self.name

    def set_value(self, value):
        msg = "Cannot specify the value of a connector '%s'"
        raise ValueError, msg % self.name

    def __str__(self):
        return self.name

    def construct(self, data=None):
        _ConnectorBase.construct(self, data)


class Connector(Component):
    """A 'bundle' of variables that can be manipulated together"""

    alias( "Connector", 
           "A bundle of variables that can be manipilated together." )

    @classmethod
    def conserved_quantity():
        pass

    # The idea behind a Connector is to create a bundle of variables
    # that can be manipulated as a single variable within constraints.
    # While Connectors inherit from variable (mostly so that the
    # expression infrastucture can manipulate them), they are not actual
    # variables that are exposed to the solver.  Instead, a preprocessor
    # (ConnectorExpander) will look for expressions that involve
    # connectors and replace the single constraint with a list of
    # constraints that involve the original variables contained within
    # the Connector.

    def __new__(cls, *args, **kwds):
        if args == ():
            self = _ConnectorElement(*args, **kwds)
        else:
            self = _ConnectorArray(*args, **kwds)
        return self




class ConnectorExpander(Plugin):
    implements(IPyomoScriptModifyInstance)

    def apply(self, **kwds):
        logger.debug("Calling ConnectorExpander")
                
        instance = kwds['instance']
        blockList = instance.all_blocks()
        noConnectors = True
        for b in blockList:
            if b.components.components(Connector):
                noConnectors = False
                break
        if noConnectors:
            return

        logger.debug("   Connectors found!")

        #
        # At this point, there are connectors in the model, so we must
        # look for constraints that involve connectors and expand them.
        #
        #options = kwds['options']
        #model = kwds['model']


        # Expand each constraint involving a connector
        for block in blockList:
            logger.debug("   block: " + block.name)
            for name, constraint in \
                    block.components.components(Constraint).iteritems():
                for idx, c in constraint._data.iteritems():
                    logger.debug("   (looking at constraint %s[%s])", name, idx)
                    connectors = []
                    self._gather_connectors(c.body, connectors)
                    if len(connectors) == 0:
                        continue
                    logger.debug("   (found connectors in constraint)")
                    
                    # Validate that all connectors match
                    errors = self._validate_connectors(connectors)
                    if errors:
                        logger.error(
                            ( "Connector mismatch: errors detected when "
                              "constructing constraint %s\n    " %
                              (name + (idx and '[%s]' % idx or '')) ) +
                            '\n    '.join(reversed(errors)) )
                        raise ValueError(
                            "Connector mismatch in constraint %s" % \
                            name + (idx and '[%s]' % idx or ''))
                    
                    logger.debug("   (connectors valid)")
                    
                    # OK - expand this constraint
                    self._expand_constraint(block, name, idx, c, connectors[0])
                    # Now deactivate the original constraint
                    c.deactivate()

        # Now, go back and implement VarList aggregators
        for block in blockList:
            for conn in block.components.components(Connector).itervalues():
                for var, aggregator in conn.aggregators.iteritems():
                    c = Constraint(expr=aggregator(block, var))
                    block._add_component(
                        conn.name + '.' + var.name + '.aggregate', c)
                    c.construct()
        

        # REQUIRED: re-call preprocess()
        instance.preprocess()

    def _gather_connectors(self, expr, connectors):
        if expr.is_expression():
            if expr.__class__ is _ProductExpression:
                for e in expr._numerator:
                    self._gather_connectors(e, connectors)
                for e in expr._denominator:
                    self._gather_connectors(e, connectors)
            else:
                for e in expr._args:
                    self._gather_connectors(e, connectors)
        elif isinstance(expr, _ConnectorValue):
            connectors.append(expr)

    def _validate_connectors(self, connectors):
        ref = connectors.pop()
        errors = []
        for tmp in connectors:
            a = sorted(ref.vars.keys())
            b = sorted(tmp.vars.keys())
            while a:
                if not b:
                    break
                if a[-1] == b[-1]:
                    #if len(ref.vars[a[-1]]) != len(tmp.vars[a[-1]]):
                    #    errors.append(
                    #        "Variable dimension mismatch for variable "
                    #        "'%s' in connectors '%s' and '%s'" %
                    #        ( a[-1], ref.name, tmp.name ) )
                    a.pop()
                    b.pop()
                elif a[-1] > b[-1]:
                    # TODO: add a fq_name so we can easily get
                    # the full model.block.connector name
                    errors.append(
                        "Connector '%s' missing variable '%s' "
                        "(appearing in reference connector '%s')" %
                        ( tmp.name, a[-1], ref.name ) )
                    a.pop()
                else:
                    errors.append(
                        "Reference connector '%s' missing variable '%s' "
                            "(appearing in connector '%s')" %
                        ( ref.name, b[-1], tmp.name ) )
                    b.pop()
            for x in reversed(a):
                errors.append(
                    "Reference connector '%s' missing variable '%s' "
                    "(appearing in connector '%s')" %
                    ( tmp.name, x, ref.name ) )
            for x in reversed(b):
                errors.append(
                    "Connector '%s' missing variable '%s' "
                    "(appearing in reference connector '%s')" %
                    ( ref.name, x, tmp.name ) )
        return errors

    def _expand_constraint(self, block, name, idx, constraint, ref):
        cList = ConstraintList()
        block._add_component( name+( idx and '.'+str(idx) or '' )+'.expanded',
                              cList )

        def _substitute_vars(args, var):
            for idx, arg in enumerate(args):
                if arg.is_expression():
                    if arg.__class__ is _ProductExpression:
                        _substitute_vars(arg._numerator, var)
                        _substitute_vars(arg._denominator, var)
                    else:
                        _substitute_vars(arg._args, var)
                elif isinstance(arg, _ConnectorValue):
                    tmp = [ arg.vars[var] ]
                    _substitute_vars(tmp, var)
                    args[idx] = tmp[0]
                elif arg.__class__ is VarList:
                    args[idx] = arg.add()

        for var in ref.vars.iterkeys():
            c = [ constraint.body.clone() ]
            _substitute_vars(c, var)
            if constraint._equality:
                cList.add( ( c[0], constraint.upper ) )
            else:
                cList.add( ( constraint.lower, c[0], constraint.upper ) )

        cList.construct()

transform = ConnectorExpander()



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

from pyutilib.enum import Enum
import pyutilib.component.core
from plugin import *

def predefined_sets():
    import set_types
    ans = []
    for item in set_types._virtual_sets:
        ans.append( (item.name,item.doc) )
    return ans


def model_components():
    return [(name,ModelComponentFactory.doc(name)) for name in ModelComponentFactory.services()]



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________


import argparse
from coopr.opt.base import SolverFactory
from coopr.opt.parallel import SolverManagerFactory
from pyutilib.misc import Options, Container

try:
    from pympler import muppy
    from pympler.muppy import summary
    from pympler import tracker
    from pympler.asizeof import *
    pympler_available = True
except ImportError:
    pympler_available = False

import util


def add_model_group(parser):
    group = parser.add_argument_group('Model Options')

    group.add_argument('--preprocess', '--load',
        help='Specify a Python module that gets immediately executed (before '\
             'the optimization model is setup).  If this option is specified '\
             'multiple times, then the modules are executed in the specified '\
             'order.',
        action='append',
        dest='preprocess',
        default=[])
    group.add_argument('--model-name',
        help='The name of the model object that is created in the specified.' \
             'Pyomo module',
        action='store',
        dest='model_name',
        default='model')
    group.add_argument('--model-options',
        help='Options passed into a create_model() function to construct the '\
             'model.',
        action='append',
        dest='model_options',
        default=[])
    group.add_argument("--flatten-expressions", "--linearize-expressions",
        help="EXPERIMENTAL: An option intended for use on linear or mixed-integer models " \
             "in which expression trees in a model (constraints or objectives) are compacted " \
             "into a more memory-efficient and concise form. The expression trees themselves are eliminated. ",
        action="store_true",
        dest="linearize_expressions",
        default=False)
    group.add_argument("--skip-canonical-repn",
            help="Do not create the canonical representation. This is not necessary for solvers (e.g., ASL-based) that do not require it.",
            action="store_true",
            dest="skip_canonical_repn",
            default=False)
    group.add_argument('--save-model',
        help='Specify the filename to which the model is saved.  The suffix ' \
             'of this filename specifies the file format.  If debugging is '  \
             "on, then this defaults to writing the file 'unknown.lp'.",
        action='store',
        dest='save_model',
        default=None)
    group.add_argument('--symbolic-solver-labels',
        help='When interfacing with the solver, use symbol names derived from the model. For example, \"my_special_variable[1_2_3]\" instead of \"v1\". Useful for debugging. When using the ASL interface (--solver-io=nl), generates corresponding .row (constraints) and .col (variables) files. The ordering in these files provides a mapping from ASL index to symbolic model names.',
        action='store_true',
        dest='symbolic_solver_labels',
        default=False)
    group.add_argument('--ns', '--namespace',
        help='Specify a namespace that is used to select data in Pyomo data files.  If this is specified multiple '\
            'times then the union of the data in these namespaces is used to create the model instance.',
        action='append',
        dest='namespaces',
        default=[])
    return group


def add_logging_group(parser):
    group = parser.add_argument_group('Logging Options')

    group.add_argument('-q','--quiet',
        help='Disable all log messages except for those that refer to errors.',
        action='store_true',
        dest='quiet',
        default=False)
    group.add_argument('-w','--warning',
        help='Enable warning log messages for coopr and pyutilib.',
        action='store_true',
        dest='warning',
        default=False)
    group.add_argument('-i','--info',
        help='Enable informative log messages for coopr and pyutilib.',
        action='store_true',
        dest='info',
        default=False)
    group.add_argument('-v','--verbose',
        help="Indicate that debugging log messages should be printed.  This option can be specified multiple times to add log messages for other parts of coopr and pyutilib.",
        action='count',
        dest='verbose',
        default=0)
    group.add_argument('-d', '--debug',
        help='This option indicates that debugging is performed.  This implies the verbose flag, but it also allows exceptions to trigger a failure in which the program stack is printed.',
        action='store_true',
        dest='debug',
        default=False)
    #group.add_argument('--logfile',
        #help='Print log messages to the specified file.',
        #action='store',
        #dest='logfile',
        #default=None)
    return group


def add_solver_group(parser):
    solver_help=\
        "This option specifies the type of solver that is used "\
        "to optimize the Pyomo model instance.  Use the --help-solvers "\
        "option to get detailed information concerning how solvers are "\
        "executed."
#"This option can specify the name"\
#        "of an optimizer on the user's path, The following solver "\
#        "types are are currently supported:"
#    solver_list = SolverFactory.services()
#    solver_list = sorted( filter(lambda x: '_' != x[0], solver_list) )
#    solver_help += " %s." % ', '.join(solver_list)

    parser.add_argument('--help-solvers',
        help='Print information about the solvers can be used to solve Pyomo models.',
        action='store_true',
        dest='help_solvers',
        default=False)

    group = parser.add_argument_group('Solver Options')

    group.add_argument('--solver',
        help=solver_help,
        action='store',
        dest='solver',
        #choices=solver_list,
        default='glpk')
    group.add_argument('--solver-io',
        help='The type of IO used to execute the solver.  Different solvers support different types of IO, but the following are common options: lp - generate LP files, nl - generate NL files, python - direct Python interface, os - generate OSiL XML files.',
        action='store',
        dest='solver_io',
        #choices=['lp', 'nl', 'python', 'os'],
        default=None)
    group.add_argument('--solver-manager',
        help='Specify the technique that is used to manage solver executions.',
        action='store',
        dest='smanager_type',
        #type='choice',
        choices=SolverManagerFactory.services(),
        default='serial')
    group.add_argument('--solver-options',
        help='Options passed into the solver.',
        action='append',
        dest='solver_options',
        default=[])
    group.add_argument('--solver-suffixes',
        help='One or more solution suffixes to be extracted by the solver (e.g., rc, dual, or slack). Multiple options are specified by supplying the keyword options multiple times.',
        action='append',
        dest='solver_suffixes',
        default=[])
    group.add_argument('--timelimit',
        help='Limit to the number of seconds that the solver is run.',
        action='store',
        dest='timelimit',
        type=int,
        default=0)
    group.add_argument('--stream-solver',
        help='Stream the solver output to provide information about the '     \
             "solver's progress.",
        action='store_true',
        dest='tee',
        default=False)
    return group


def add_postsolve_group(parser):
    group = parser.add_argument_group('Post-Solve Options')

    group.add_argument('--postprocess',
        help='Specify a Python module that gets executed after optimization. ' \
             'If this option is specified multiple times, then the modules '  \
             'are executed in the specified order.',
        action='append',
        dest='postprocess',
        default=[])
    group.add_argument('-l','--log',
        help='Print the solver logfile after performing optimization.',
        action='store_true',
        dest='log',
        default=False)
    group.add_argument('--save-results',
        help='Specify the filename to which the results are saved.',
        action='store',
        dest='save_results',
        default=None)
    group.add_argument('--show-results',
        help='Print the results object after optimization.',
        action='store_true',
        dest='show_results',
        default=False)
    group.add_argument('--json',
        help='Print results in JSON format.  The default is YAML, if the PyYAML utility is available',
        action='store_true',
        dest='json',
        default=False)
    group.add_argument('-s','--summary',
        help='Summarize the final solution after performing optimization.',
        action='store_true',
        dest='summary',
        default=False)
    return group


def add_misc_group(parser):
    parser.add_argument('--help-components',
        help='Print information about modeling components supported by Pyomo',
        action='store_true',
        dest='help_components',
        default=False)

    group = parser.add_argument_group('Miscellaneous Options')

    group.add_argument('--output',
        help='Redirect output to the specified file.',
        action='store',
        dest='output',
        default=None)
    group.add_argument('-c', '--catch-errors',
        help='This option allows exceptions to trigger a failure in which the program stack is printed.',
        action='store_true',
        dest='catch',
        default=False)
    #group.add_argument('--logfile',
    group.add_argument('--disable-gc',
        help='Disable the garbage collecter.',
        action='store_true',
        dest='disable_gc',
        default=False)
    group.add_argument('--interactive',
        help='After executing Pyomo, launch an interactive Python shell.  If IPython is installed, this shell is an IPython shell.',
        action='store_true',
        dest='interactive',
        default=False)
    group.add_argument('-k','--keepfiles',
        help='Keep temporary files.',
        action='store_true',
        dest='keepfiles',
        default=False)
    group.add_argument('--path',
        help='Give a path that is used to find the Pyomo python files.',
        action='store',
        dest='path',
        default='.')
    group.add_argument('--profile',
        help='Enable profiling of Python code.  The value of this option is ' \
             'the number of functions that are summarized.',
        action='store',
        dest='profile',
        type=int,
        default=0)
    if pympler_available is True:
        group.add_argument("--profile-memory",
                             help="If Pympler is available, report memory usage statistics for the generated instance and any associated processing steps. A value of 0 indicates disabled. A value of 1 forces the print of the total memory after major stages of the pyomo script. A value of 2 forces summary memory statistics after major stages of the pyomo script. A value of 3 forces detailed memory statistics during instance creation and various steps of preprocessing. Values equal to 4 and higher currently provide no additional information. Higher values automatically enable all functionality associated with lower values, e.g., 3 turns on detailed and summary statistics.",
                             action="store",
                             dest="profile_memory",
                             type=int,
                             default=0)
    group.add_argument('--report-timing',
        help='Report various timing statistics during model construction. Defaults to disabled.',
        action='store_true',
        dest='report_timing',
        default=False)
    group.add_argument('--tempdir',
        help='Specify the directory where temporary files are generated.',
        action='store',
        dest='tempdir',
        default=None)
    return group


def create_parser():
    #
    #
    # Setup command-line options
    #
    #
    parser = argparse.ArgumentParser(
                usage = '%(prog)s [options] <model_file> [<data_files>]'
                )
    group = add_model_group(parser)
    group.add_argument('--instance-only',
        help='Generate a model instance, and then return.',
        action='store_true',
        dest='only_instance',
        default=False)
    add_solver_group(parser)
    add_postsolve_group(parser)
    add_logging_group(parser)
    add_misc_group(parser)
    parser.add_argument('model_file', action='store', nargs='?', default='', help='A Python module that defines a Pyomo model')
    parser.add_argument('data_files', action='store', nargs='*', default=[], help='Pyomo data files that defined data used to create a model instance')
    return parser


def run_pyomo(options=Options(), parser=None):
    data = Options(options=options)
    if options.help_solvers:
        util.print_solver_help(data)
        return Container()
    #
    if options.help_components:
        util.print_components(data)
        return Container()
    #
    util.setup_environment(data)
    #
    util.apply_preprocessing(data, parser=parser)
    if data.error:
        return Container()                                   #pragma:nocover
    #
    model_data = util.create_model(data)
    if (not options.debug and options.save_model) or options.only_instance:
        return Container(instance=model_data.instance)
    #
    opt_data = util.apply_optimizer(data, instance=model_data.instance)

    # this is hack-ish, and carries the following justification.
    # symbol maps are not pickle'able, and as a consequence, results
    # coming back from a pyro solver manager don't have a symbol map.
    # however, you need a symbol map to load the result into an 
    # instance. so, if it isn't there, construct it!
    if opt_data.results._symbol_map is None:
        from coopr.opt.base.symbol_map import symbol_map_from_instance
        results._symbol_map = symbol_map_from_instance(model_data.instance)

    #
    util.process_results(data, instance=model_data.instance, results=opt_data.results, opt=opt_data.opt)
    #return Container(instance=model_data.instance, results=results) #pragma:nocover
    #
    util.apply_postprocessing(data, instance=model_data.instance, results=opt_data.results)
    #
    util.finalize(data, model=model_data.model, instance=model_data.instance, results=opt_data.results)
    #
    return Container(options=options, instance=model_data.instance, results=opt_data.results)


def run(args=None):
    result = util.run_command(command=run_pyomo, parser=create_parser(), args=args, name='pyomo')
    return result.retval

def main(args=None):
    ans = run(args)
    if ans is None:
        return 0
    return 1




"""
       / -9x-20, -5 <= x <= -4
       | -7x-12, -4 <= x <= -3
       | -5x-6 , -3 <= x <= -2
       | -3x-2 , -2 <= x <= -1
f(x) = | 1     , -1 <= x <=  1 
       | 3x-2  ,  1 <= x <=  2
       | 5x-6  ,  2 <= x <=  3
       \ 7x-12 ,  3 <= x <=  4

"""

from coopr.pyomo import *

DOMAIN_PTS = [float(i) for i in (range(-5,0)+range(1,5))]

def F(model,x):
    return x**2

def define_model(**kwds):

    model = ConcreteModel()

    model.x1 = Var(bounds=(-5,4)) # domain variable
    model.x2 = Var(bounds=(-5,4)) # domain variable    
    model.x3 = Var(bounds=(-5,4)) # domain variable
    model.x4 = Var(bounds=(-5,4)) # domain variable
    model.x5 = Var(bounds=(-5,4)) # domain variable
    model.x6 = Var(bounds=(-5,4)) # domain variable
    model.x7 = Var(bounds=(-5,4)) # domain variable

    model.Fx1 = Var() # range variable
    model.Fx2 = Var() # range variable
    model.Fx3 = Var() # range variable
    model.Fx4 = Var() # range variable
    model.Fx5 = Var() # range variable
    model.Fx6 = Var() # range variable
    model.Fx7 = Var() # range variable

    model.obj = Objective(expr=model.Fx1+model.Fx2+model.Fx3+model.Fx4+model.Fx5+model.Fx6+model.Fx7, sense=kwds.pop('sense',maximize))

    model.piecewise1 = Piecewise(model.Fx1,model.x1,
                                  pw_pts=DOMAIN_PTS,
                                  f_rule=F, **kwds)
    model.piecewise2 = Piecewise(model.Fx2,model.x2,
                                  pw_pts=DOMAIN_PTS,
                                  f_rule=F, **kwds)
    model.piecewise3 = Piecewise(model.Fx3,model.x3,
                                  pw_pts=DOMAIN_PTS,
                                  f_rule=F, **kwds)
    model.piecewise4 = Piecewise(model.Fx4,model.x4,
                                  pw_pts=DOMAIN_PTS,
                                  f_rule=F, **kwds)
    model.piecewise5 = Piecewise(model.Fx5,model.x5,
                                  pw_pts=DOMAIN_PTS,
                                  f_rule=F, **kwds)
    model.piecewise6 = Piecewise(model.Fx6,model.x6,
                                  pw_pts=DOMAIN_PTS,
                                  f_rule=F, **kwds)
    model.piecewise7 = Piecewise(model.Fx7,model.x7,
                                  pw_pts=DOMAIN_PTS,
                                  f_rule=F, **kwds)

    #Fix the answer for testing purposes
    model.set_answer_constraint1 = Constraint(expr= model.x1 == -5.0)
    model.set_answer_constraint2 = Constraint(expr= model.x2 == -3.0)
    model.set_answer_constraint3 = Constraint(expr= model.x3 == -2.5)
    model.set_answer_constraint4 = Constraint(expr= model.x4 == -1.5)
    model.set_answer_constraint5 = Constraint(expr= model.x5 == 2.0)
    model.set_answer_constraint6 = Constraint(expr= model.x6 == 3.5)
    model.set_answer_constraint7 = Constraint(expr= model.x7 == 4.0)    
    
    return model



#
# Unit Tests for nontrivial Bounds (_SumExpression, _ProductExpression)
#

import os
import sys
from os.path import abspath, dirname
sys.path.insert(0, dirname(dirname(abspath(__file__)))+"/../..")
currdir = dirname(abspath(__file__))+os.sep

import pyutilib.th as unittest
from coopr.pyomo import *
from nose.tools import nottest

class Test(unittest.TestCase):

    #Test constraint bounds
    def test_constr_lower(self):
        self.model = AbstractModel()
        self.model.A = Param(default=2.0, mutable=True)
        self.model.B = Param(default=1.5, mutable=True)
        self.model.C = Param(default=2.5, mutable=True)
        self.model.X = Var()

        def constr_rule(model):
            return (self.model.A*(self.model.B+self.model.C),self.model.X)
        self.model.constr = Constraint(rule=constr_rule)

        self.instance = self.model.create()
        self.instance.pprint()
        self.assertEqual(self.instance.constr.lower(),8.0)

    def test_constr_upper(self):
        self.model = AbstractModel()
        self.model.A = Param(default=2.0, mutable=True)
        self.model.B = Param(default=1.5, mutable=True)
        self.model.C = Param(default=2.5, mutable=True)
        self.model.X = Var()

        def constr_rule(model):
            return (self.model.X,self.model.A*(self.model.B+self.model.C))
        self.model.constr = Constraint(rule=constr_rule)

        self.instance = self.model.create()

        self.assertEqual(self.instance.constr.upper(),8.0)

    def test_constr_both(self):
        self.model = AbstractModel()
        self.model.A = Param(default=2.0, mutable=True)
        self.model.B = Param(default=1.5, mutable=True)
        self.model.C = Param(default=2.5, mutable=True)
        self.model.X = Var()

        def constr_rule(model):
            return (self.model.A*(self.model.B-self.model.C),self.model.X,self.model.A*(self.model.B+self.model.C))
        self.model.constr = Constraint(rule=constr_rule)

        self.instance = self.model.create()

        self.assertEqual(self.instance.constr.lower(),-2.0)
        self.assertEqual(self.instance.constr.upper(),8.0)


    #Test variable bounds
    #JPW: Disabled until we are convinced that we want to support complex parametric expressions for variable bounds.
    def test_var_bounds(self):
        self.model = AbstractModel()
        self.model.A = Param(default=2.0, mutable=True)
        self.model.B = Param(default=1.5, mutable=True)
        self.model.C = Param(default=2.5, mutable=True)

        def X_bounds_rule(model):
            return (self.model.A*(self.model.B-self.model.C),self.model.A*(self.model.B+self.model.C))
        self.model.X = Var(bounds=X_bounds_rule)

        self.instance = self.model.create()

        self.assertEqual(self.instance.X.lb(),-2.0)
        self.assertEqual(self.instance.X.ub(),8.0)
   




from checker import *
from runner import *
from script import *
from hooks import *

# Modules
__all__ = ['checkers']

# Checker classes
__all__.extend(['IModelChecker', 'PyomoModelChecker'])
__all__.extend(['ImmediateDataChecker', 'IterativeDataChecker'])
__all__.extend(['ImmediateTreeChecker', 'IterativeTreeChecker'])

# Other builtins
__all__.extend(['ModelCheckRunner', 'ModelScript'])

# Hooks
__all__.extend(['IPreCheckHook', 'IPostCheckHook'])
